/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "DecoderBase.hpp"
#include "StagePosibility.hpp"
#include "TransitionCostComputation.hpp"
#include "Token.hpp"
#include "Splitter.hpp"


namespace ngramchecker {

	void DecoderBase::init_posibilities(vector<Token> &tokens)
	{
		stage_posibilities = StagePosibilitiesType(new vector<vector<StagePosibilityP> >());

		vector<StagePosibilityP> start_sp;
		start_sp.push_back(sentence_start_stage_posibility);

		for (uint32_t i = 0; i < viterbi_order - 1; i++)
		{
			stage_posibilities->push_back(start_sp);
		}

		vector<vector<StagePosibilityP> > inner_sp = init_inner_stage_posibilities(tokens);
		stage_posibilities->insert(stage_posibilities->end(), inner_sp.begin(), inner_sp.end());

		vector<StagePosibilityP> end_sp;
		end_sp.push_back(sentence_end_stage_posibility);

		stage_posibilities->push_back(end_sp);

	}


	uint32_t DecoderBase::GetViterbiOrder()
	{
		return viterbi_order;
	}

	TransitionCostComputationP DecoderBase::GetTransitionCostComputation()
	{
		return transition_cost_computation;
	}

	vector<StagePosibilityP>  DecoderBase::DecodeTokenizedSentence(vector<Token> &tokens)
	{

		//cerr << "decoded init...\n";

		vector<StagePosibilityP> ret_vec;

		if (tokens.size() == 0)
		{
			return ret_vec;
		}

		init_posibilities(tokens);



		Trellis_stage_setP new_trellis_stage;
		Trellis_stageP last_trellis_stage = Trellis_stageP(new Trellis_stage());


		vector<StagePosibilityP> start_state_history;

		for (uint32_t i = 0; i < viterbi_order - 1; i++)
			start_state_history.push_back(sentence_start_stage_posibility);

		ViterbiStateP init_vsP = ViterbiStateP(new ViterbiState(start_state_history));
		last_trellis_stage->push_back(init_vsP);

		//cerr << "...Decoder start...\n";


		for (uint32_t i = viterbi_order - 1; i < stage_posibilities->size(); i++)
		{
			vector<StagePosibilityP> current_stage_posibilities = stage_posibilities->operator [](i);
			new_trellis_stage = Trellis_stage_setP(new Trellis_stage_set());

			//cerr << "stage = " << i << ", possibilities = " << current_stage_posibilities.size() << endl;

			double best = 100000000;

			assert(last_trellis_stage->size() > 0);

			for (uint32_t j = 0; j < last_trellis_stage->size(); j++)
			{
				ViterbiStateP viterbi_stateP = last_trellis_stage->operator [](j);



				for (uint32_t k = 0; k < current_stage_posibilities.size(); k++)
				{
					StagePosibilityP stage_pos = current_stage_posibilities[k];

					double cost = viterbi_stateP->distance;
					double emmision_cost = stage_pos->EmmisionProbability();

					cost += emmision_cost;

					//cerr << "distance: " << viterbi_stateP->distance << ", emmision: " << emmision_cost << endl;

					assert(cost < 100000000);

					if (cost > best + MyConstants::prunning_constant)
						continue;

					double transition_cost = transition_cost_computation->ComputeTransitionCost(viterbi_stateP, stage_pos);


					if (transition_cost < 0)
					{
						cerr << "!!! transition_cost < 0 !!! == " << transition_cost << ", (might happen, because of prob. values quantizations)" <<endl;
						//exit(1);
					}

					cost += transition_cost;


					if (cost > best + MyConstants::prunning_constant)
						continue;

					if (cost < best)
					{
						//cerr << "old best = " << best << ", new best = " << cost << "(emm = " << emmision_cost << ", trans = " << transition_cost << ")" << endl;
						best = cost;
					}

					//LMStateKeyP new_lm_state_keyP = LMStateKeyP(new LMStateKey(viterbi_stateP->lmState, wf_cp.word_factors));
					//GlobalVars::lmServer->Reduce(new_lm_state_keyP);

					ViterbiStateP new_viterbi_stateP = ViterbiStateP(new ViterbiState(viterbi_stateP, stage_pos, cost));

					Trellis_stage_set::iterator it = new_trellis_stage->find(new_viterbi_stateP);

					if (it == new_trellis_stage->end())
					{
						new_trellis_stage->insert(new_viterbi_stateP);
					}
					else
					{
						ViterbiStateP stored_state = *it;
						if (stored_state->distance > new_viterbi_stateP->distance)
						{
							stored_state->distance = new_viterbi_stateP->distance;
							stored_state->ancestor = new_viterbi_stateP->ancestor;
						}

					}

				}

			}

			last_trellis_stage = Trellis_stageP(new Trellis_stage());

			for (Trellis_stage_set::iterator it = new_trellis_stage->begin(); it != new_trellis_stage->end(); it++)
			{
				ViterbiStateP state = *it;
				if (state->distance < best + MyConstants::prunning_constant)
				{
					last_trellis_stage->push_back(state);
				}
			}

			std::sort(last_trellis_stage->begin(), last_trellis_stage->end(), ViterbiStateP_compare_distance());
			//cerr << "last_trellis_stage->size() = " << last_trellis_stage->size() << endl;
		}

		assert(last_trellis_stage->size() > 0);

		ViterbiStateP state = last_trellis_stage->operator [](0);

		ret_vec.push_back(state->GetYoungestHistory());


		while (state->ancestor != NULL)
		{
			state = state->ancestor;
			ret_vec.push_back(state->GetYoungestHistory());
		}

		if (viterbi_order > 1)
		{
			for (uint32_t i = 0; i < viterbi_order - 2; i++)
				ret_vec.push_back(sentence_start_stage_posibility);
		}

		assert(ret_vec.size() == stage_posibilities->size());

		std::reverse(ret_vec.begin(), ret_vec.end());
		return ret_vec;
	}

	void DecoderBase::DecodeTokenizedSentence_ReturnStagePosibilities(vector<Token> &tokens, vector<StagePosibilityP> &decoded_sequence, StagePosibilitiesType &_stage_posibilities)
	{
		decoded_sequence = DecodeTokenizedSentence(tokens);
		_stage_posibilities = stage_posibilities;
	}

	string DecoderBase::DecodeSentence(string sentence)
	{
		vector<string> tokens_str;
		vector<string> tok_delims;

		Splitter::SplitSentenceToTokens(sentence, tokens_str, tok_delims);

		vector<Token> tokens;

		for (uint32_t i = 0; i < tokens_str.size(); i++)
			tokens.push_back(Token(tokens_str[i]));

		vector<StagePosibilityP> spv = DecodeTokenizedSentence(tokens);


		stringstream sstr;

		for (uint32_t i = viterbi_order - 1; i < spv.size() - 1; i++)
		{
			sstr << spv[i]->ToString();

			if (i - (viterbi_order - 1) < tok_delims.size())
				sstr << tok_delims[i - (viterbi_order - 1)];
		}

		string ret = sstr.str();

		return ret;
	}

}
