#ifndef _ERROR_MODEL_BASIC_HPP
#define _ERROR_MODEL_BASIC_HPP

#include "StdAfx.h"
#include "ErrorModel.hpp"
#include "utils.hpp"
#include "GlobalVars.hpp"
#include "StringMapper.hpp"

#define triple(A, B, C) pair<A, pair<B, C>>
#define make_triple(A, B, C) make_pair(A, make_pair(B, C))

#define FIRST(A) A.first
#define SECOND(A) A.second.first
#define THIRD(A) A.second.second

namespace ngramchecker {

	class ErrorModelBasic;
	SP_DEF(ErrorModelBasic);

    struct char_pair_hash : std::unary_function<pair<char, char>, size_t>
	{
        size_t operator()(const pair<char, char> &val) const
		{
			size_t seed = 0;
			boost::hash_combine(seed, val.first);
			boost::hash_combine(seed, val.second);
			return seed;
		}
	};

    struct char_triple_hash : std::unary_function<triple(char, char, char), size_t>
	{
        size_t operator()(const triple(char, char, char) &val) const
		{
			size_t seed = 0;
			boost::hash_combine(seed, FIRST(val));
			boost::hash_combine(seed, SECOND(val));
			boost::hash_combine(seed, THIRD(val));
			return seed;
		}
	};

	class ErrorModelBasic : public ErrorModel
	{
	private:

        unordered_map<pair<char, char>, ErrorModelOutput, char_pair_hash> substitution_map;
        unordered_map<triple(char, char, char), ErrorModelOutput, char_triple_hash> insertion_map;
        unordered_map<pair<char, char>, ErrorModelOutput, char_pair_hash> deletion_map;
        unordered_map<pair<char, char>, ErrorModelOutput, char_pair_hash> swap_map;

		ErrorModelOutput substitution_default;
		ErrorModelOutput insertion_default;
		ErrorModelOutput deletion_default;
		ErrorModelOutput swap_default;
		ErrorModelOutput case_mismatch_cost;
		uint min_operation_edit_distance;

        char convert(char16_t ch16)
        {
            if (ch16 == 0x161)
                return 0xB9;
            else if (ch16 == 0x165)
                return 0xBB;
            else if (ch16 == 0x17E)
                return 0xBE;
            else if (ch16 == 0x10D)
                return 0xE8;
            else if (ch16 == 0x11B)
                return 0xEC;
            else if (ch16 == 0x10F)
                return 0xEF;
            else if (ch16 == 0x148)
                return 0xF2;
            else if (ch16 == 0x159)
                return 0xF8;
            else if (ch16 == 0x16F)
                return 0xF9;
			else if (ch16 == 0x160)
				return 0xA9;
			else if (ch16 == 0x164)
				return 0xAB;
			else if (ch16 == 0x17D)
				return 0xAE;
			else if (ch16 == 0x10C)
				return 0xC8;
			else if (ch16 == 0x11A)
				return 0xCC;
			else if (ch16 == 0x10E)
				return 0xCF;
			else if (ch16 == 0x147)
				return 0xD2;
			else if (ch16 == 0x158)
				return 0xD8;
			else if (ch16 == 0x16E)
				return 0xD9;
            else if (((uint16_t)ch16) > 255)
            {
                cerr << "unicode character conversion failure! - " << ((uint16_t)ch16) << endl;
                //exit(-123123);
            }

            return (char)ch16;
        }

	public:		

		ErrorModelBasic(istream &ifs)
		{
			ifs.read((char*)&(substitution_default.cost), sizeof(float));
			ifs.read((char*)&(substitution_default.edit_dist), sizeof(uint32_t));
			ifs.read((char*)&(insertion_default.cost), sizeof(float));
			ifs.read((char*)&(insertion_default.edit_dist), sizeof(uint32_t));
			ifs.read((char*)&(deletion_default.cost), sizeof(float));
			ifs.read((char*)&(deletion_default.edit_dist), sizeof(uint32_t));
			ifs.read((char*)&(swap_default.cost), sizeof(float));
			ifs.read((char*)&(swap_default.edit_dist), sizeof(uint32_t));
			ifs.read((char*)&(case_mismatch_cost.cost), sizeof(float));
			ifs.read((char*)&(case_mismatch_cost.edit_dist), sizeof(uint32_t));
			ifs.read((char*)&min_operation_edit_distance, sizeof(uint32_t));

            char16_t ch1;
            char16_t ch2;
            char16_t ch3;
			float cost;
			uint32_t ed_dist;

			uint num_subs;
			ifs.read((char*)&num_subs, sizeof(uint32_t));

			for (uint i = 0; i < num_subs; i++)
			{
                ifs.read((char*)&ch1, sizeof(char16_t));
                ifs.read((char*)&ch2, sizeof(char16_t));
                ifs.read((char*)&cost, sizeof(float));
				ifs.read((char*)&ed_dist, sizeof(uint32_t));

                substitution_map[make_pair(convert(ch1), convert(ch2))] = ErrorModelOutput(ed_dist, cost);
			}

			uint num_inserts;
			ifs.read((char*)&num_inserts, sizeof(uint32_t));

			for (uint i = 0; i < num_inserts; i++)
			{
                ifs.read((char*)&ch1, sizeof(char16_t));
                ifs.read((char*)&ch2, sizeof(char16_t));
                ifs.read((char*)&ch3, sizeof(char16_t));
				ifs.read((char*)&cost, sizeof(float));
				ifs.read((char*)&ed_dist, sizeof(uint32_t));

                insertion_map[make_triple(convert(ch1), convert(ch2), convert(ch3))] = ErrorModelOutput(ed_dist, cost);
			}

			uint num_dels;
			ifs.read((char*)&num_dels, sizeof(uint32_t));

			for (uint i = 0; i < num_dels; i++)
			{
                ifs.read((char*)&ch1, sizeof(char16_t));
                ifs.read((char*)&ch2, sizeof(char16_t));
				ifs.read((char*)&cost, sizeof(float));
				ifs.read((char*)&ed_dist, sizeof(uint32_t));

                deletion_map[make_pair(convert(ch1), convert(ch2))] = ErrorModelOutput(ed_dist, cost);
			}

			uint num_swaps;
			ifs.read((char*)&num_swaps, sizeof(uint32_t));

			for (uint i = 0; i < num_swaps; i++)
			{
                ifs.read((char*)&ch1, sizeof(char16_t));
                ifs.read((char*)&ch2, sizeof(char16_t));
				ifs.read((char*)&cost, sizeof(float));
				ifs.read((char*)&ed_dist, sizeof(uint32_t));

                swap_map[make_pair(convert(ch1), convert(ch2))] = ErrorModelOutput(ed_dist, cost);
			}

		}

		static ErrorModelBasicP fromBinaryFile(string binary_file)
		{
			ifstream ifs;
			ifs.open(binary_file.c_str(), ios::binary);
			FATAL_CONDITION(ifs.is_open(), "");
			
			ErrorModelBasicP ret = ErrorModelBasicP(new ErrorModelBasic(ifs));
			ifs.close();
			return ret;
		}

        virtual ErrorModelOutput SubstitutionCost(char char1, char char2, bool ignore_case = false)
		{

			if (ignore_case == false)
			{
				auto fit = substitution_map.find(make_pair(char1, char2));
				if (fit == substitution_map.end())
				{
					if (char1 == char2)
						return ErrorModelOutput(0, 0.0f);
					else 
					{
						if (GlobalVars::stringMapper->CharIsUpperCase(char1))
							char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char1);

						if (GlobalVars::stringMapper->CharIsUpperCase(char2))
							char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char2);

						if (char1 == char2)
							return case_mismatch_cost;
						else
							return substitution_default;
					}
				}
				else
					return fit->second;
			}
			else
			{
				if (GlobalVars::stringMapper->CharIsUpperCase(char1))
					char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char1);

				if (GlobalVars::stringMapper->CharIsUpperCase(char2))
					char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char2);

				auto fit = substitution_map.find(make_pair(char1, char2));
				if (fit == substitution_map.end())
				{
					if (char1 == char2)
						return ErrorModelOutput(0, 0.0f);
					else
						return substitution_default;
				}
				else
					return fit->second;

			}
		}


		//yep, this looks really weird, yet, it's correct
        virtual ErrorModelOutput DeletionCost(char current_char, char previous_char, char next_char)
		{
			auto fit = insertion_map.find(make_triple(current_char, previous_char, next_char));
			if (fit == insertion_map.end())
				return insertion_default;
			else
				return fit->second;
		}

        virtual ErrorModelOutput SwapCost(char first_char, char second_char)
		{
			auto fit = swap_map.find(make_pair(first_char, second_char));

			if (fit == swap_map.end())
				return swap_default;
			else
				return fit->second;
		}

        virtual ErrorModelOutput InsertionCost(char inserted_char, char previous_char, char next_char)
		{
			auto fit = deletion_map.find(make_pair(inserted_char, previous_char));
			if (fit == deletion_map.end())
				return deletion_default;
			else
				return fit->second;
		}

		virtual uint32_t MinOperationEditDistance() { return min_operation_edit_distance; }

        virtual bool StringsAreIdentical(const string &s1, const string &s2) { return s1 == s2; }

		virtual string ToString() { return "ErrorModelBasic"; }

	    virtual double UnknownWordCost()
		{
			return ErrorModel::impossible_cost;
		}

		virtual double NameCost()
		{
			return ErrorModel::impossible_cost;
		}

	};

}

#endif //_ERROR_MODEL_BASIC_HPP
