/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "ErrorModelTypo.hpp"
#include "GlobalVars.hpp"
#include "StringMapper.hpp"

namespace ngramchecker {

	    double ErrorModelTypo::UnknownWordCost()
		{
			return ErrorModel::impossible_cost;
		}

		double ErrorModelTypo::NameCost()
		{
			return ErrorModel::impossible_cost;
		}

		//letter1 ... original letter
		//letter2 ... possible correction
		ErrorModelOutput ErrorModelTypo::SubstitutionCost(char char1, char char2, bool ignore_case)
		{
			
			if (ignore_case == false)
			{
				double cap_cost = 0;
				
				if (GlobalVars::stringMapper->CharIsUpperCase(char1) && GlobalVars::stringMapper->CharIsLowerCase(char2))
				{
					char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char1);
					cap_cost = cost_sub_capitalization;
				}
				else if (GlobalVars::stringMapper->CharIsUpperCase(char2) && GlobalVars::stringMapper->CharIsLowerCase(char1))
				{
					char2 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char2);
					cap_cost = cost_sub_capitalization;				
				}
				
				char_pair chp = pair<char, char>(char1, char2);

				if (char1 == char2)
					return ErrorModelOutput(0, cap_cost);

				map<char_pair, double>::iterator fit = substitution_cost_map.find(chp);

				if (fit != substitution_cost_map.end())
				{
					return ErrorModelOutput(1, fit->second + cap_cost);
				}
				else
					return ErrorModelOutput(1, cost_sub_random + cap_cost);
			}
			else
			{
				if (GlobalVars::stringMapper->CharIsUpperCase(char1))
				{
					char1 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char1);
				}
				if (GlobalVars::stringMapper->CharIsUpperCase(char2))
				{
					char2 = GlobalVars::stringMapper->GetLowerCaseForUpperCase_char(char2);
				}

				return SubstitutionCost(char1, char2, false);
			}

		}

		//TODO: remove parameters
		ErrorModelOutput ErrorModelTypo::InsertionCost(char char_inserted, char previous_char, char next_char)
		{
				return ErrorModelOutput(1, cost_insert);
		}

		ErrorModelOutput ErrorModelTypo::SwapCost(char first_char, char second_char)
		{
			return ErrorModelOutput(1, cost_swap);
		}

		ErrorModelOutput ErrorModelTypo::DeletionCost(char current_char, char previous_char, char next_char)
		{
			if (hadj_letters.find(char_pair(current_char, previous_char)) != hadj_letters.end() ||
				hadj_letters.find(char_pair(current_char, next_char)) != hadj_letters.end())
			{
				return ErrorModelOutput(1, cost_delete_hadj);
			}
			else if (current_char == previous_char || current_char == next_char)
			{
				return ErrorModelOutput(1, cost_delete_same);
			}
			else if (vadj_letters.find(char_pair(current_char, previous_char)) != vadj_letters.end() ||
				vadj_letters.find(char_pair(current_char, next_char)) != vadj_letters.end())
			{
				return ErrorModelOutput(1, cost_delete_vadj);
			}
			else
			{
				return ErrorModelOutput(1, cost_delete_random);
			}
		}

		uint32_t ErrorModelTypo::MinOperationEditDistance()
		{
			return 1;
		}

		bool ErrorModelTypo::StringsAreIdentical(const string &s1, const string &s2)
		{
			return s1 == s2;
		}

		string ErrorModelTypo::ToString()
		{
			return "typo";
		}

		ErrorModelTypo::ErrorModelTypo(string error_model_file)
		{
			string s;

			ifstream cost_stream;
			cost_stream.open( error_model_file.c_str() );

			while (std::getline(cost_stream, s))
			{
				if (s == "cor-insert")
				{
					std::getline(cost_stream, s);

					cost_insert = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-del:hadj")
				{
					std::getline(cost_stream, s);

					cost_delete_hadj = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-del:random")
				{
					std::getline(cost_stream, s);

					cost_delete_random = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-del:same")
				{
					std::getline(cost_stream, s);

					cost_delete_same = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-del:vadj")
				{
					std::getline(cost_stream, s);

					cost_delete_vadj = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:diac_add")
				{
					std::getline(cost_stream, s);

					cost_sub_diac_del = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:diac_del")
				{
					std::getline(cost_stream, s);

					cost_sub_diac_add = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:hadj")
				{
					std::getline(cost_stream, s);

					cost_sub_hadj = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:iy")
				{
					std::getline(cost_stream, s);

					cost_sub_iy = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:random")
				{
					std::getline(cost_stream, s);

					cost_sub_random = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:sz")
				{
					std::getline(cost_stream, s);

					cost_sub_sz = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:vadj")
				{
					std::getline(cost_stream, s);

					cost_sub_vadj = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:vocal")
				{
					std::getline(cost_stream, s);

					cost_sub_vocal = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:yi")
				{
					std::getline(cost_stream, s);

					cost_sub_yi = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-sub:zs")
				{
					std::getline(cost_stream, s);

					cost_sub_zs = boost::lexical_cast<double>(s);
					continue;
				}

				if (s == "cor-swap")
				{
					std::getline(cost_stream, s);

					cost_swap = boost::lexical_cast<double>(s);
					continue;
				}
				
				if (s == "cor-capitalization")
				{
					std::getline(cost_stream, s);
					
					cost_sub_capitalization = boost::lexical_cast<double>(s);
					continue;
				}

				cerr << "Line wasn't processed! : " << s << endl;

			}

			ifstream istr;
			istr.open( MyConstants::filename_keyboard_layout.c_str() );
			assert(istr.is_open());


			std::getline(istr, s);
			assert(s == "===vocal");

			while (true) //vocal substitution section
			{
				std::getline(istr, s);

				if (s == "===del_diac") break;

				if (vocal_sub_map.find(s[0]) == vocal_sub_map.end())
					vocal_sub_map[s[0]] = vector<char>();

				if (vocal_sub_map.find(s[1]) == vocal_sub_map.end())
					vocal_sub_map[s[1]] = vector<char>();

				vocal_sub_map[s[0]].push_back(s[1]);
				vocal_sub_map[s[1]].push_back(s[0]);

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);
				char_pair char_pair2 = char_pair(char2, char1);


				substitution_cost_map[char_pair1] = cost_sub_vocal;
				substitution_cost_map[char_pair2] = cost_sub_vocal;
			}

			while (true) //del_diac section
			{
				std::getline(istr, s);


				if (s == "===add_diac") break;

				del_diac_map[s[0]] = s[1];

				//cerr << s << endl;
				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);

				substitution_cost_map[char_pair1] = cost_sub_diac_del;

			}

			//cerr << "Add diac section!" << endl;
			while (true) //add_diac section
			{
				std::getline(istr, s);

				if (s == "===vadj") break;

				if (add_diac_map.find(s[0]) == add_diac_map.end())
				{
					add_diac_map[s[0]] = vector<char>();
				}

				add_diac_map[s[0]].push_back(s[1]);

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);

				substitution_cost_map[char_pair1] = cost_sub_diac_add;

			}

			//cerr << "vadj section!!!" << endl;
			while (true) //vadj section
			{
				std::getline(istr, s);

				//cerr << s << endl;
				if (s == "===hadj") break;

				if (vadj_map.find(s[0]) == vadj_map.end())
					vadj_map[s[0]] = vector<char>();

				if (vadj_map.find(s[1]) == vadj_map.end())
					vadj_map[s[1]] = vector<char>();

				vadj_map[s[0]].push_back(s[1]);
				vadj_map[s[1]].push_back(s[0]);

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);
				char_pair char_pair2 = char_pair(char2, char1);

				substitution_cost_map[char_pair1] = cost_sub_vadj;
				substitution_cost_map[char_pair2] = cost_sub_vadj;

				vadj_letters.insert(char_pair1);
				vadj_letters.insert(char_pair2);

			}

			//cerr << "hadj section!!!" << endl;
			while (true) //hadj section
			{

				std::getline(istr, s);

				if (s == "===yi") break;

				if (hadj_map.find(s[0]) == hadj_map.end())
					hadj_map[s[0]] = vector<char>();

				if (hadj_map.find(s[1]) == hadj_map.end())
					hadj_map[s[1]] = vector<char>();

				hadj_map[s[0]].push_back(s[1]);
				hadj_map[s[1]].push_back(s[0]);

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);
				char_pair char_pair2 = char_pair(char2, char1);

				substitution_cost_map[char_pair1] = cost_sub_hadj;
				substitution_cost_map[char_pair2] = cost_sub_hadj;

				hadj_letters.insert(char_pair1);
				hadj_letters.insert(char_pair2);

			}

			while (true) //yi
			{
				std::getline(istr, s);

				if (s == "===iy") break;

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);

				substitution_cost_map[char_pair1] = cost_sub_yi;

			}

			while (std::getline(istr, s)) //iy
			{

				char char1 = s[0];
				char char2 = s[1];

				char_pair char_pair1 = char_pair(char1, char2);

				substitution_cost_map[char_pair1] = cost_sub_iy;

			}

			char letter_s = 's';
			char letter_z = 'z';
			char letter_S = 'S';
			char letter_Z = 'Z';

			substitution_cost_map[char_pair(letter_s, letter_z)] = cost_sub_sz;
			substitution_cost_map[char_pair(letter_z, letter_s)] = cost_sub_zs;
			substitution_cost_map[char_pair(letter_S, letter_Z)] = cost_sub_sz;
			substitution_cost_map[char_pair(letter_Z, letter_S)] = cost_sub_zs;

			istr.close();

			ifstream letters_stream;
			letters_stream.open(MyConstants::filename_lc_uc_letters.c_str());
			assert(letters_stream.is_open());

			std::getline(letters_stream, lc_chars);
			std::getline(letters_stream, uc_chars);
			std::getline(letters_stream, lc_chars_no_diacritics);
			std::getline(letters_stream, uc_chars_no_diacritics);

			letters_stream.close();
		}

		void ErrorModelTypo::PrintCosts()
		{
			cerr << "cost_sub_hadj: " << cost_sub_hadj << endl;
			cerr << "cost_sub_random: " << cost_sub_random << endl;
			cerr << "cost_sub_vadj: " << cost_sub_vadj << endl;
			cerr << "cost_sub_zs: " << cost_sub_zs << endl; //z is error, s is true
			cerr << "cost_sub_vocal: " << cost_sub_vocal << endl;
			cerr << "cost_sub_diac_del: " << cost_sub_diac_del << endl;
			cerr << "cost_sub_diac_add: " << cost_sub_diac_add << endl;
			cerr << "cost_sub_yi: " << cost_sub_yi << endl; //y is error, i is true
			cerr << "cost_sub_sz: " << cost_sub_sz << endl;  //s is error, z is true
			cerr << "cost_sub_iy: " << cost_sub_iy << endl; //i is error, y is true
			cerr << "cost_sub_capitalization: " << cost_sub_capitalization << endl;

			cerr << "cost_delete_random: " << cost_delete_random << endl;
			cerr << "cost_delete_same: " << cost_delete_same << endl;
			cerr << "cost_delete_hadj: " << cost_delete_hadj << endl;
			cerr << "cost_delete_vadj: " << cost_delete_vadj << endl;

			cerr << "cost_insert: " << cost_insert << endl;

			cerr << "cost_swap: " << cost_swap << endl;

		}


}
