/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "Evaluation.hpp"
#include "GlobalVars.hpp"
#include "utils.hpp"
#include "SimWordsFinder.hpp"
#include "SimWordsFinderDiacritic.hpp"
#include "SimWordsFinderSpelling.hpp"
#include "ErrorModel.hpp"
#include "ErrorModelTypo.hpp"
#include "DecoderFormsOnly.hpp"
#include "DecoderMultiFactor.hpp"
#include "Spellchecker.hpp"
#include "DiacriticComplement.hpp"
#include "Tagger.hpp"


namespace ngramchecker {

	string Evaluation::EvaluateCorrectionFormsOnly(string out_dir, string dataset, bool ignore_case, double _form_weight, uint32_t _form_order)
	{
		string out_file = MyConstants::directory_evaluation_correction + "exp/" + out_dir + "/output.txt";
		string desc_file = MyConstants::directory_evaluation_correction + "exp/" + out_dir + "/describtion.info";

		ofstream desc_ofs;
		desc_ofs.open(desc_file.c_str(), ios::out );
		assert(desc_ofs.is_open());

		desc_ofs << "time: " << MyUtils::GetTimeString() << endl << "decoder: Forms Only" << endl << "dataset: " << dataset << endl
		<< "ignore case: " << MyUtils::BoolToString(ignore_case) << endl
				 << "form_weight: " << _form_weight << endl
				 << "form_order: " << _form_order << endl;


		desc_ofs.close();

		ifstream ifs;
		ifs.open( (MyConstants::directory_evaluation_correction + dataset + ".dirty").c_str(), ios::in );
		assert(ifs.is_open());

		ofstream ofs;
		ofs.open( out_file.c_str(), ios::out );
		assert(ofs.is_open());

		string s;
		uint32_t counter = 0;

		SimWordsFinderP swf = SimWordsFinderSpelling::CreateWithDefaultParams(GlobalVars::errorModelTypoP, ErrorModelP(), ignore_case);

		while (std::getline(ifs, s))
		{

			if (counter > 0) { ofs << endl; }

			if (s == "") { counter++; continue; }

			DecoderBaseP dfo = DecoderBaseP(new DecoderFormsOnly(_form_order, swf, _form_weight));

			Spellchecker spellchecker = Spellchecker(dfo);

			string output = spellchecker.DecodeEvaluation(s, 5);
			ofs << output;

			counter++;
			if (counter % 20 == 0)
				cerr << counter << endl;
		}

		ifs.close();
		ofs.close();

		return out_file;
	}

	void Evaluation::EvaluateCorrectionMultiFactor(string out_dir, string dataset, double form_weight, double lemma_weight, double tag_weight, bool ignore_case, uint32_t form_lm_order, uint32_t lemma_lm_order, uint32_t tag_lm_order)
	{

		string out_file = MyConstants::directory_evaluation_correction + "exp/" + out_dir + "/output.txt";
		string desc_file = MyConstants::directory_evaluation_correction + "exp/" + out_dir + "/describtion.info";

		ofstream desc_ofs;
		desc_ofs.open(desc_file.c_str(), ios::out );
		assert(desc_ofs.is_open());

		desc_ofs << "time: " << MyUtils::GetTimeString() << endl << "decoder: Forms Only" << endl << "dataset: " << dataset << endl
		<< "ignore case: " << MyUtils::BoolToString(ignore_case) << endl << endl << "form weight: " << form_weight << endl
				 << "lemma weight: " << lemma_weight << endl << "tag weight: " << tag_weight << endl;

		desc_ofs.close();

		ifstream ifs;
		ifs.open( (MyConstants::directory_evaluation_correction + dataset + ".dirty").c_str(), ios::in );
		assert(ifs.is_open());

		ofstream ofs;
		ofs.open( out_file.c_str(), ios::out );
		assert(ofs.is_open());

		string s;
		uint32_t counter = 0;

		SimWordsFinderP swf = SimWordsFinderSpelling::CreateWithDefaultParams(GlobalVars::errorModelTypoP, ErrorModelP(), ignore_case);

		while (std::getline(ifs, s))
		{
			if (s == "") continue;
			if (counter > 0) { ofs << endl; }

			DecoderBaseP dmf = DecoderBaseP(new DecoderMultiFactor(form_lm_order, lemma_lm_order, tag_lm_order, swf, form_weight, lemma_weight, tag_weight));

			Spellchecker spellchecker = Spellchecker(dmf);

			string output = spellchecker.DecodeEvaluation(s, 5);

			ofs << output;

			counter++;
			if (counter % 20 == 0)
				cerr << counter << endl;
		}

		ifs.close();
		ofs.close();

	}

	void Evaluation::EvaluateDiacriticFormsOnly(string out_dir, string dataset, ErrorModelP error_model, bool ignore_case, LetterLMSettings llms, double _form_weight, uint32_t _form_order)
	{
		//assert(error_model == GlobalVars::errorModelDiacriticsSimpleP || error_model == GlobalVars::errorModelDiacriticsComplexP);


		string out_file = MyConstants::directory_evaluation_diacritic + "exp/" + out_dir + "/output.txt";
		string desc_file = MyConstants::directory_evaluation_diacritic + "exp/" + out_dir + "/describtion.info";

		ofstream desc_ofs;
		desc_ofs.open(desc_file.c_str(), ios::out );
		assert(desc_ofs.is_open());

		desc_ofs << "time: " << MyUtils::GetTimeString() << endl << "decoder: Forms Only" << endl << "dataset: " << dataset
		<< "error model: " << error_model->ToString() << endl << "ignore case: " << MyUtils::BoolToString(ignore_case) << endl
		<< "use_letter_lm: " << MyUtils::BoolToString(llms.use_letter_lm) << endl << "letter_lm_change_normal_cost: " << llms.change_cost_normal << endl
				 << "letter_lm_change_name_cost: " << llms.change_cost_name << endl
				 << "form_weight: " << _form_weight << endl
				 << "form_order: " << _form_order << endl;

		desc_ofs.close();

		ofstream ofs;
		ofs.open( out_file.c_str(), ios::out );
		assert(ofs.is_open());

		ifstream ifs;
		ifs.open( (MyConstants::directory_evaluation_diacritic + dataset + ".dirty").c_str(), ios::in );
		assert(ifs.is_open());

		string s;
		uint32_t counter = 0;

		SimWordsFinderP swf = SimWordsFinderDiacriticP(new SimWordsFinderDiacritic(error_model, ignore_case, llms));

		while (std::getline(ifs, s))
		{
			if (counter > 0) { ofs << endl; }

			DecoderBaseP dfo = DecoderBaseP(new DecoderFormsOnly(_form_order, swf, _form_weight));

			DiacriticComplement dc = DiacriticComplement(dfo);

			//cerr << ">>>>" << endl;

			string output = dc.AddDiacriticToText(s);
			ofs << output;

			//cerr << "<<<<" << endl;

			counter++;
			if (counter % 20 == 0)
				cerr << counter << endl;
		}

		ifs.close();
		ofs.close();

	}

	string Evaluation::EvaluateDiacriticMultiFactor(string out_dir, string dataset, ErrorModelP error_model, double form_weight, double lemma_weight, double tag_weight, bool ignore_case, LetterLMSettings llms, uint32_t form_lm_order, uint32_t lemma_lm_order, uint32_t tag_lm_order)
	{
		//assert(error_model == GlobalVars::errorModelDiacriticsSimpleP || error_model == GlobalVars::errorModelDiacriticsComplexP);

		string out_file = MyConstants::directory_evaluation_diacritic + "exp/" + out_dir + "/output.txt";
		string desc_file = MyConstants::directory_evaluation_diacritic + "exp/" + out_dir + "/describtion.info";

		ofstream desc_ofs;
		desc_ofs.open(desc_file.c_str(), ios::out );
		assert(desc_ofs.is_open());

		desc_ofs << "time: " << MyUtils::GetTimeString() << endl << "decoder: MultiFactor" << endl << "dataset: " << dataset << endl
		<< "error model: " << error_model->ToString() << endl << "ignore case: " << MyUtils::BoolToString(ignore_case) << endl
		<< "use letter lm: " << MyUtils::BoolToString(llms.use_letter_lm) << endl << "letter_lm_change_normal_cost: " << llms.change_cost_normal << endl
				 << "letter_lm_change_name_cost: " << llms.change_cost_name << endl << endl
				 << "form weight: " << form_weight << endl << "lemma weight: " << lemma_weight << endl << "tag weight: " << tag_weight << endl
				 << "form order: " << form_lm_order << endl << "lemma order: " << lemma_lm_order << endl << "tag order: " << tag_lm_order << endl;

		desc_ofs.close();

		ofstream ofs;
		ofs.open( out_file.c_str(), ios::out );
		assert(ofs.is_open());

		ifstream ifs;
		ifs.open( (MyConstants::directory_evaluation_diacritic + dataset + ".dirty").c_str(), ios::in );
		assert(ifs.is_open());

		string s;
		uint32_t counter = 0;

		SimWordsFinderP swf = SimWordsFinderDiacriticP(new SimWordsFinderDiacritic(error_model, ignore_case, llms));

		while (std::getline(ifs, s))
		{
			if (counter > 0) { ofs << endl; }

			DecoderBaseP dmf = DecoderBaseP(new DecoderMultiFactor(form_lm_order, lemma_lm_order, tag_lm_order, swf, form_weight, lemma_weight, tag_weight));

			DiacriticComplement dc = DiacriticComplement(dmf);

			string output = dc.AddDiacriticToText(s);

			ofs << output;

			counter++;
			if (counter % 20 == 0)
				cerr << counter << endl;
		}

		ifs.close();
		ofs.close();

		return out_file;
	}

	string Evaluation::EvaluateTagger(string dataset)
	{

		cerr << "Evaluate tagger: Need to be reimplemented!";
		exit(1);

		ifstream ifs;
		ifs.open( (MyConstants::directory_evaluation_tagger + dataset + ".form").c_str(), ios::in );
		assert(ifs.is_open());

		ofstream ofs;

		stringstream of_strs;
		of_strs << MyConstants::directory_evaluation_tagger << "exp/" << MyUtils::GetTimeString() <<"-tagger-" << dataset;
		string out_file = of_strs.str();


		ofs.open( out_file.c_str(), ios::out );
		assert(ofs.is_open());

		string s;
		uint32_t counter = 0;

		while (std::getline(ifs, s))
		{
			if (counter > 0) { ofs << endl; }

			Tagger tagger = Tagger(3);

			string output = tagger.TagSentence(s);
			ofs << output;

			counter++;
			if (counter % 20 == 0)
				cerr << counter << endl;
		}

		ifs.close();
		ofs.close();

		return out_file;
	}

}