/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "Lexicon.hpp"
#include "GlobalVars.hpp"
#include "StringMapper.hpp"
#include "MorphologyMapper.hpp"
#include "LemmaSigns.hpp"
#include "utils.hpp"
#include "MyPackedArray.hpp"
#include "CompIncreasingArray.hpp"
#include "ErrorModel.hpp"

namespace ngramchecker {

	  boost::shared_ptr<set<LexiconArc> > Lexicon::Arcs_set(uint32_t stateID)
	  {
		  boost::shared_ptr<set<LexiconArc> > ret = boost::shared_ptr<set<LexiconArc> >(new set<LexiconArc>());

		  if (stateID < states_arcpointer->GetSize())
		  {
			  CompIA_First_Last_IndexPair index_pair = states_arcpointer->GetFirstLastIndexPair(stateID);

			  if (index_pair == NULL)
				  return ret;

			  for (uint32_t i = index_pair->first; i <= index_pair->second; i++)
			  {
				  ret->insert(LexiconArc(GlobalVars::stringMapper->InvMapChar(arcs_isymbol->GetValueAt(i)), arcs_nextstate->GetValueAt(i)));
			  }
		  }

		  return ret;
	  }

	  vectorP_LexiconArc Lexicon::Arcs_vector(uint32_t stateID)
	  {
		  if (lexicon_cash.IsCashed(stateID))
		  {
			  return lexicon_cash.GetCashedValue(stateID);
		  }

		  vectorP_LexiconArc ret = boost::shared_ptr<vector<LexiconArc> >(new vector<LexiconArc>());

		  if (stateID < states_arcpointer->GetSize())
		  {
			  CompIA_First_Last_IndexPair index_pair = states_arcpointer->GetFirstLastIndexPair(stateID);

			  if (index_pair == NULL)
			  {
				  lexicon_cash.StoreValueForKey(stateID, ret);
				  return ret;
			  }

			  for (uint32_t i = index_pair->first; i <= index_pair->second; i++)
			  {
				  ret->push_back(LexiconArc(GlobalVars::stringMapper->InvMapChar(arcs_isymbol->GetValueAt(i)), arcs_nextstate->GetValueAt(i)));
			  }
		  }

		  lexicon_cash.StoreValueForKey(stateID, ret);
		  return ret;
	  }

	  uint32_t Lexicon::GetWordID_per_state_id(uint32_t state_ID)
	  {
		  if (state_ID >= n_inner_states)
			  return state_ID - n_inner_states + GlobalVars::morphologyMapper->GetFirstNonSpecialID();
		  else
		  {
			  return GlobalVars::morphologyMapper->GetUnkTokenID();
		  }
	  }

	  uint32_t Lexicon::NumStates() const { return states_arcpointer->GetSize(); }


	  int Lexicon::ContainsSuffix(uint32_t stateID, const string &word, uint32_t start_char_index, bool check_lemma_sign)
	  {
		  int sID = (int)stateID;
		  for (uint32_t i = start_char_index; i < word.length(); i++)
		  {

			 sID = SingleArc_nextstate((uint32_t)sID, word[i]);
			 if (sID == -1)
				 return -1;
		  }

		  if ( StateIsWord( (uint32_t)sID, check_lemma_sign ) )
		  {
			return  GetWordID_per_state_id((uint32_t)sID);
		  }
		  else
		  {
			  return -1;
		  }
	  }



	  bool Lexicon::StateIsWord(uint32_t state_ID, bool check_lemma_sign)
	  {
		  bool ret = (state_ID >= n_inner_states && state_ID < states_arcpointer->GetSize());

		  if (ret == false || check_lemma_sign == false)
			  return ret;
		  else
		  {
			  uint32_t word_id = GetWordID_per_state_id(state_ID);

			  return GlobalVars::lemmaSigns->IsFormOkey(word_id);
		  }
	  }



	  int Lexicon::SingleArc_nextstate(uint32_t stateID, char character)
	  {
		  uint32_t isymbol = GlobalVars::stringMapper->MapChar(character);
		  CompIA_First_Last_IndexPair index_pair = states_arcpointer->GetFirstLastIndexPair(stateID);

		  if (index_pair == NULL)
			  return -1;

		  uint32_t leftIndex = index_pair->first;
		  uint32_t rightIndex = index_pair->second;
		  uint32_t actIndex;

		  do {
			  actIndex = (leftIndex + rightIndex) >> 1;
			  uint32_t arc_isymbol = arcs_isymbol->GetValueAt(actIndex);

			  if (arc_isymbol == isymbol )
			  {
				  return arcs_nextstate->GetValueAt(actIndex);
			  }

			  if (arc_isymbol > isymbol)
			  {
				  if (rightIndex == actIndex)
					  return -1;
				  rightIndex = actIndex;
			  }
			  else
			  {

				  leftIndex = actIndex;

				  if (leftIndex == actIndex && rightIndex == leftIndex + 1)
				  {
					  leftIndex = rightIndex;
					  continue;
				  }
			  }

			  if (leftIndex == rightIndex)
				  break;
			  //cerr << leftIndex << " " << rightIndex << endl;
		  } while (true);

		  return -1;
	  }

	  Lexicon* Lexicon::ReadFromBinFile(string filename)
	  {
		  ifstream ifs;
		  uint32_t _n_inner_states;
		  ifs.open(filename.c_str(), ios::binary | ios::in);

		  assert(ifs.is_open());

		  ifs.read((char*)&_n_inner_states, sizeof(uint32_t));

		  boost::shared_ptr<CompIncreasingArray> states_arcpointer = boost::shared_ptr<CompIncreasingArray>(new CompIncreasingArray(ifs));
		  boost::shared_ptr<MyPackedArray> arcs_nexts = boost::shared_ptr<MyPackedArray>(new MyPackedArray(ifs));
		  boost::shared_ptr<MyPackedArray> arcs_isym = boost::shared_ptr<MyPackedArray>(new MyPackedArray(ifs));

		  ifs.close();
		  return new Lexicon(_n_inner_states, states_arcpointer, arcs_nexts, arcs_isym);
	  }

	  void Lexicon::WriteToBinFile(string filename)
	  {
		  ofstream ofs;
		  ofs.open(filename.c_str(), ios::binary | ios::out);

		  assert(ofs.is_open());

		  ofs.write((char*)&n_inner_states, sizeof(uint32_t));
		  states_arcpointer->WriteToStream(ofs);
		  arcs_nextstate->WriteToStream(ofs);
		  arcs_isymbol->WriteToStream(ofs);

		  ofs.close();
	  }

	  int Lexicon::GetWordID(const string &word_iso) //return value -1 for out of vocabulary words!
	  {
		  if (word_iso.length() == 1 && GlobalVars::stringMapper->CharIsPunctuation(word_iso[0]))
		  {
			  return GlobalVars::morphologyMapper->PunctuationUniversalIndex_char(word_iso[0]);
		  }
		  else if (GlobalVars::stringMapper->WordIsNumber(word_iso))
		  {
			  return GlobalVars::morphologyMapper->GetNumberTokenID();
		  }
		  else
		  {
			  return ContainsSuffix(0, word_iso, 0, false);
		  }
	  }

	  Lexicon::Lexicon(map<uint32_t, map<uint32_t, uint32_t> > states_map, uint32_t _n_inner_states):
		  lexicon_cash(MyConstants::default_lexicon_state_cash_capacity)
	  {

		n_inner_states = _n_inner_states;
		vector<uint32_t> vstates_arcpointer;
		vector<uint32_t> varcs_nextstate;
		vector<uint32_t> varcs_isymbol;

		cerr << "n_inner_states = " << n_inner_states << endl;

		uint32_t pos = 0;
		uint32_t s = 0;
		for (map<uint32_t, map<uint32_t, uint32_t> >::iterator sIt = states_map.begin(); sIt != states_map.end(); sIt++)
		{
		  //cerr << "s == " << s << ", sIt->first == " << sIt->first << endl;
		  if (! (s == 0 || s == sIt->first - 1) )
		  {
			  cerr << "s == " << s << ", sIt->first == " << sIt->first << endl;
			  exit(1);
		  }
		  s = sIt->first;
		  vstates_arcpointer.push_back(pos);
		  map<uint32_t, uint32_t> arcmap = states_map[s];

		  for (map<uint32_t, uint32_t>::iterator aiter = arcmap.begin(); aiter != arcmap.end(); aiter++)
		  {
			varcs_nextstate.push_back(aiter->second);
			varcs_isymbol.push_back(aiter->first);
			pos++;
		  }
		}

		arcs_nextstate = boost::shared_ptr<MyPackedArray>(new MyPackedArray(varcs_nextstate));
		arcs_isymbol = boost::shared_ptr<MyPackedArray>(new MyPackedArray(varcs_isymbol));
		states_arcpointer = boost::shared_ptr<CompIncreasingArray>(new CompIncreasingArray(vstates_arcpointer, arcs_nextstate->GetSize() - 1));

		for (uint32_t i = 0; i < states_arcpointer->GetSize(); i++)
		{
			boost::shared_ptr<set<LexiconArc> > arcs = Arcs_set(i);
			for (set<LexiconArc>::iterator it = arcs->begin(); it != arcs->end(); it++)
			{
				LexiconArc arc = *it;
				assert(states_map[i].find(GlobalVars::stringMapper->MapChar(arc.character)) != states_map[i].end());
				assert(states_map[i][GlobalVars::stringMapper->MapChar(arc.character)] == arc.nextstate);
			}
		}

		cerr << "Test inside the Lexicon constructor passed!" << endl;

		/*for (map<uint32_t, map<uint32_t, uint32_t> >::iterator sIt = states_map.begin(); iIt != states_map.end(); sIt++)
		{
			  for (map<uint32_t, uint32_t>::iterator aiter = arcmap.begin(); aiter != arcmap.end(); aiter++)

		}*/
	  }



	  void Lexicon::TestIt()
	  {
		  vector<uint32_t> acumStates;
		  acumStates.push_back(0);
		  vector< boost::shared_ptr<vector<LexiconArc> > > acumArcs;
		  acumArcs.push_back( Arcs_vector(0) );
		  vector<uint32_t> acumArcPointers;
		  acumArcPointers.push_back(0);

		  uint32_t stateID;
		  uint32_t arcPointer;

		  map<uint32_t, string> stored_words;
		  string word = "";


		  while (acumStates.size() > 0)
		  {
			  assert(acumStates.size() == acumArcPointers.size() && acumStates.size() == acumArcs.size());

			  stateID = acumStates[acumStates.size() - 1];
			  arcPointer = acumArcPointers[acumArcPointers.size() - 1];
			  boost::shared_ptr<vector<LexiconArc> > arcs = acumArcs[acumArcs.size() - 1];

			  if (arcPointer == arcs->size())
			  {
				  //cerr << "StateIsWord(" << stateID << " ~ " << word << ") n_inner_states = " << n_inner_states << ", n_states = " << states_arcpointer->GetSize() << endl;
				  if (StateIsWord(stateID, false))
				  {
					  stored_words[GetWordID_per_state_id(stateID)] = word;
					 //cerr << "stored_words[" << (GetWordID_per_state_id(stateID)) << "] = " << word << endl;
				  }

				  acumStates.pop_back();
				  acumArcs.pop_back();
				  acumArcPointers.pop_back();
				  word = word.substr(0, word.length() - 1);
			  }
			  else
			  {
				  LexiconArc la = arcs->operator [](arcPointer);
				  word += la.character;
				  acumStates.push_back(la.nextstate);
				  acumArcPointers[acumArcPointers.size() - 1]++;
				  acumArcPointers.push_back(0);
				  acumArcs.push_back( Arcs_vector(la.nextstate) );
			  }

		  }

		  cerr << "Lexicon crawled!\n";

		  for (uint32_t i = GlobalVars::morphologyMapper->GetFirstNonSpecialID(); i < GlobalVars::stringMapper->NumForms(); i++)
		  {
			  if (GlobalVars::stringMapper->InvMapForm(i) != stored_words[i])
			  {
				  cerr << "Error: GlobalVars::stringMapper->InvMapForm(" << i << ") = "
					  << GlobalVars::stringMapper->ConvertBackFromISO(GlobalVars::stringMapper->InvMapForm(i)) <<
						  ", stored_words[" << i << "] = " << stored_words[i] << endl;
			  }
			  assert(GlobalVars::stringMapper->InvMapForm(i) == stored_words[i]);
		  }

		  cerr << "Lexicon test passed!";
	  }

	  void Lexicon::AddSimilarWordToMap(Similar_Words_Map &ret, uint32_t word_id, double cost, string &word, uint32_t word_include_letter_start_index, string &prefix)
	  {
		  //bool ret_val = false;
		  Similar_Words_Map::iterator it = ret.find(word_id);
		  if (it == ret.end())
		  {
			  string complete_word = prefix;
			  for (uint32_t i = word_include_letter_start_index; i < word.length(); i++)
				  complete_word.push_back(word[i]);

			  ret[word_id] = pair<stringP, double>(stringP(new string(complete_word)), cost);
		  }
		  else if (it->second.second > cost)
		  {
			  ret[word_id].second = cost;
		  }

		  //return ret_val;
	  }


	  Similar_Words_Map Lexicon::GetSimilarWords_impl(string &word, uint32_t edit_distance, double cost, ErrorModelP errModel, uint32_t stateID, uint32_t startIndex, string &prefix, double cost_limit, bool ignore_case, uint32_t max_edit_distance)
	  {
		  //cerr << edit_distance << endl;
		  Similar_Words_Map ret;
		  double new_cost;
		  uint32_t new_edit_dist;
		  ErrorModelOutput emo;

		  uint32_t init_prefix_size = prefix.size();


		  for (uint32_t i = startIndex; i <= word.length(); i++)
		  {
			  if (i > startIndex && i <= word.length()) //TODO: Check this
				  prefix.push_back(word[i - 1]);

			  if (i == word.length() && StateIsWord(stateID, false))
			  {
				 AddSimilarWordToMap(ret, GetWordID_per_state_id(stateID), cost, word, i, prefix);
			  }

			  int next_state = -1;
			  boost::shared_ptr<vector<LexiconArc> > arcs = Arcs_vector(stateID);
		 	  if (i < word.size())
			  {

		 		  if (word.size() > 1)
		 		  {
					  if (i == 0)
						 emo = errModel->DeletionCost(word[i], word[i + 1], word[i + 1]);
					  else if (i == word.size() - 1)
						 emo = errModel->DeletionCost(word[i], word[i - 1], word[i - 1]);
					  else
						 emo = errModel->DeletionCost(word[i], word[i - 1], word[i + 1]);

					  new_cost = cost + emo.cost;
					  new_edit_dist = edit_distance + emo.edit_dist;

					  if (new_cost <= cost_limit && new_edit_dist <= max_edit_distance)
					  {
						  int wID = ContainsSuffix(stateID, word, i + 1, false);
						  if (wID != -1)
							AddSimilarWordToMap(ret, wID, new_cost, word, i + 1, prefix);

						  if (new_edit_dist + errModel->MinOperationEditDistance() <= max_edit_distance)
						  {
							  Similar_Words_Map sim_w = GetSimilarWords_impl(word, new_edit_dist, new_cost, errModel, stateID, i + 1, prefix, cost_limit, ignore_case, max_edit_distance);
							  ret.insert(sim_w.begin(), sim_w.end());
						  }

					  }
		 		  }

				  for (uint32_t j = 0; j < arcs->size(); j++) //letter substitution
				  {
					  LexiconArc arc = arcs->operator [](j);
					  if (arc.character != word[i])
					  {
						  emo = errModel->SubstitutionCost(word[i], arc.character, ignore_case);
						  new_cost = cost + emo.cost;
						  new_edit_dist = edit_distance + emo.edit_dist;

						  if (new_cost <= cost_limit && new_edit_dist <= max_edit_distance)
						  {

							  int wID = ContainsSuffix(arc.nextstate, word, i + 1, false);

							  prefix.push_back(arc.character);

							  if (wID != -1)
								  AddSimilarWordToMap(ret, wID, new_cost, word, i + 1, prefix);

							  if (new_edit_dist + errModel->MinOperationEditDistance() <= max_edit_distance)
							  {
								  Similar_Words_Map sim_w = GetSimilarWords_impl(word, new_edit_dist, new_cost, errModel, arc.nextstate, i + 1, prefix, cost_limit, ignore_case, max_edit_distance);
								  ret.insert(sim_w.begin(), sim_w.end());
							  }

							  prefix.erase(prefix.size() - 1);
						  }

					  }
					  else
					  {
						  next_state = (int)arc.nextstate;
					  }
				  }

				  if (i < word.size() - 1) //letter swap
				  {

					  emo = errModel->SwapCost(word[i], word[i + 1]);
					  new_cost = cost + emo.cost;
					  new_edit_dist = edit_distance + emo.edit_dist;

					  if (new_cost <= cost_limit && new_edit_dist <= max_edit_distance)
					  {
						  char pom_swap = word[i];
						  word[i] = word[i + 1];
						  word[i + 1] = pom_swap;


						  int wID = ContainsSuffix(stateID, word, i, false);
						  if (wID != -1)
							  AddSimilarWordToMap(ret, wID, new_cost, word, i, prefix);

						  if (new_edit_dist + errModel->MinOperationEditDistance() <= max_edit_distance)
						  {
							  Similar_Words_Map sim_w = GetSimilarWords_impl(word, new_edit_dist, new_cost, errModel, stateID, i, prefix, cost_limit, ignore_case, max_edit_distance);
							  ret.insert(sim_w.begin(), sim_w.end());
						  }


						  word[i + 1] = word[i];
						  word[i] = pom_swap;
					  }
				  }

			  } // if (i < word.size() )

			  for (uint32_t j = 0; j < arcs->size(); j++) //letter insertion
			  {
				  LexiconArc arc = arcs->operator [](j);

				  if (i == 0)
					  emo = errModel->InsertionCost(arc.character, word[0], word[0]);
				  else if (i == word.size())
					  emo = errModel->InsertionCost(arc.character, word[i - 1], word[i - 1]);
				  else
					  emo = errModel->InsertionCost(arc.character, word[i - 1], word[i]);

				  new_cost = cost + emo.cost;
				  new_edit_dist = edit_distance + emo.edit_dist;

				  if (new_cost <= cost_limit && new_edit_dist <= max_edit_distance)
				  {
					  prefix.push_back(arc.character);

					  int wID = ContainsSuffix(arc.nextstate, word, i, false);
					  if (wID != -1)
					  {
						AddSimilarWordToMap(ret, wID, new_cost, word, i, prefix);
					  }

					  if (new_edit_dist + errModel->MinOperationEditDistance() <= max_edit_distance)
					  {
						  Similar_Words_Map sim_w = GetSimilarWords_impl(word, new_edit_dist, new_cost, errModel, arc.nextstate, i, prefix, cost_limit, ignore_case, max_edit_distance);
						  ret.insert(sim_w.begin(), sim_w.end());
					  }

					  prefix.erase(prefix.size() - 1);
				  }




			  }

		 	  if (next_state == -1)
		 		  break;
		 	  stateID = next_state;

		  }

		  while (prefix.size() > init_prefix_size)
			  prefix.erase(prefix.size() - 1);

		  return ret;
	  }

	  Similar_Words_Map Lexicon::GetSimilarWords(string word, uint32_t max_ed_dist, double _cost_limit, ErrorModelP errModel, bool _ignore_case)
	  {
		  string prefix = "";
		  return GetSimilarWords_impl(word, 0, 0, errModel, 0, 0, prefix, _cost_limit, _ignore_case, max_ed_dist);
	  }

	  vector<string> Lexicon::GetSimilarWords_string(string word, uint32_t max_ed_dist, double _cost_limit, ErrorModelP errModel, bool _ignore_case)
	  {
		vector<string> ret;

		Similar_Words_Map swm = GetSimilarWords(word, max_ed_dist, _cost_limit, errModel, _ignore_case);

		for (Similar_Words_Map::iterator it = swm.begin(); it != swm.end(); it++)
		{
			ret.push_back(*(it->second.first));
		}

		return ret;
	  }

}

