#ifndef MORPHOLOGY_HPP_
#define MORPHOLOGY_HPP_

#include "StdAfx.h"
#include "MyPackedArray.hpp"
#include "CompIncreasingArray.hpp"
#include "MyBitArray.hpp"
#include "MyStaticStringArray.h"

namespace ngramchecker {

	class morpho_grouping {
		MyPackedArrayP group_values;
		CompIncreasingArrayP group_offsets;

	public:
		vector<uint> getGroupMembers(uint groupID)
		{
			vector<uint> ret;
			CompIA_First_Last_IndexPair index_pair = group_offsets->GetFirstLastIndexPair(groupID);
			for (uint i = index_pair->first; i <= index_pair->second; i++)
				ret.push_back(group_values->GetValueAt(i));
			return ret;
		}

		morpho_grouping(const MyPackedArrayP &_group_values, const CompIncreasingArrayP &_group_offsets):
			group_values(_group_values), group_offsets(_group_offsets) {}
	};

	SP_DEF(morpho_grouping);

	class morpho_dependency {
		uint governing_factor;
		MyPackedArrayP dependency_mapping;
	public:
		uint GetValue(uint governing_factor_id)
		{
			return dependency_mapping->GetValueAt(governing_factor_id);
		}

		inline uint GetGoverningFactorIndex()
		{
			return governing_factor;
		}

		morpho_dependency(uint _governing_factor, const MyPackedArrayP &dep_mapping): governing_factor(_governing_factor), dependency_mapping(dep_mapping) {}
	};

	SP_DEF(morpho_dependency);

	class Morphology {

		uint num_factors;
		MyStaticStringArrayP factor_names;
		vector<uint> bits_per_value; //number of bits needed to store a factorID for the particular factor
		vector<uint> bits_per_children; //number of bits needed to store a number of children for a node at the particular level or a groupID if the level is grouped
		vector<morpho_dependencyP> dependencies; //list of all factor dependencies
		vector<morpho_groupingP> groupings;

		CompIncreasingArrayP formOffsets;
		MyBitArrayP morphoData;
		vector<MyStaticStringArrayP> morpho_dictionary;

	private:

		void get_morphology_rec(uint level, vector<uint> &curr_factors, vector<vector<uint> > &ret, uint& bit_offset, int next_factor = -1)
		{
			if (next_factor >= 0)
				curr_factors.push_back((uint)next_factor);
			else if (dependencies[level])
			{
				uint cur_fact = dependencies[level]->GetValue(curr_factors[dependencies[level]->GetGoverningFactorIndex()]);
				curr_factors.push_back(cur_fact);
			}
			else if (level > 0)
			{
				uint cur_fact = morphoData->GetValueAt(bit_offset, bits_per_value[level]);
				curr_factors.push_back(cur_fact);
				bit_offset += bits_per_value[level];
			}

			if (level < num_factors - 1)
			{
				if (!(dependencies[level + 1]))
				{
					
					if (groupings[level + 1])
					{
						uint groupID = morphoData->GetValueAt(bit_offset, bits_per_children[level]);
						bit_offset += bits_per_children[level];

						vector<uint> group_members = groupings[level + 1]->getGroupMembers(groupID);
						for (uint i = 0; i < group_members.size(); i++)
							get_morphology_rec(level + 1, curr_factors, ret, bit_offset, group_members[i]);
					}
					else
					{
						uint num_children = morphoData->GetValueAt(bit_offset, bits_per_children[level]);
						num_children++; //there is never zero children so the values are shifted
						bit_offset += bits_per_children[level];

						for (uint i = 0; i < num_children; i++)
							get_morphology_rec(level + 1, curr_factors, ret, bit_offset);
					}
				}
				else
				{
					get_morphology_rec(level + 1, curr_factors, ret, bit_offset);
				}

			}
			else
			{
				ret.push_back(curr_factors);
			}

			curr_factors.pop_back();
		}

	public:
		
		void initMorphoDictionary(string filename)
		{
			morpho_dictionary.clear();
			ifstream ifs;
			ifs.open(filename.c_str(), ios::in | ios::binary);
			assert(ifs.is_open());

			uint num_f;
			ifs.read((char*)&num_f, sizeof(uint32_t));
			assert(num_f == num_factors);

			for (uint i = 0; i < num_factors; i++)
			{
				morpho_dictionary.push_back(MyStaticStringArrayP(new MyStaticStringArray(ifs)));
			}
		}

		vector<vector<uint> > GetMorphology(uint form_id)
		{
			vector<vector<uint> > ret;
			uint bit_offset = formOffsets->GetValueAt(form_id);

			vector<uint> curr_factors;
			curr_factors.push_back(form_id);

			get_morphology_rec(0, curr_factors, ret, bit_offset);

			return ret;
		}

		void PrintOut(ofstream &ofs)
		{
			assert(morpho_dictionary.size() > 0);

			for (uint i = 0; i < formOffsets->GetSize(); i++)
			{
				vector<vector<uint> > morpho_anal = GetMorphology(i);

				for (uint j = 0; j < morpho_anal.size(); j++)
				{
					for (uint k = 0; k < num_factors; k++)
					{
						if (k > 0) ofs << "|";
						ofs << morpho_dictionary[k]->GetStringAt(morpho_anal[j][k]);
					}
					ofs << endl;
				}
			}
		}


		Morphology(string &filename)
		{
			ifstream ifs;
			ifs.open(filename.c_str(), ios::in | ios::binary);
			if (!ifs.is_open())
			{
				cerr << "Can't open " << filename << std::endl;
				exit(1);
			}
			
			ifs.read((char*)&num_factors, sizeof(uint32_t));

			for (size_t i = 0; i < num_factors; i++)
			{
				dependencies.push_back(morpho_dependencyP());
				groupings.push_back(morpho_groupingP());
			}

			factor_names = MyStaticStringArrayP(new MyStaticStringArray(ifs));

			uint num_dependencies;
			ifs.read((char*)&num_dependencies, sizeof(uint32_t));

			assert(num_dependencies < num_factors);

			for (uint i = 0; i < num_dependencies; i++)
			{
				uint governing;
				uint governed;
				ifs.read((char*)&governing, sizeof(uint32_t));
				ifs.read((char*)&governed, sizeof(uint32_t));

				assert(governing < governed);

				MyPackedArrayP dep_array = MyPackedArrayP(new MyPackedArray(ifs));

				dependencies[governed] = morpho_dependencyP(new morpho_dependency(governing, dep_array));
			}
	
			uint num_grouped;
			ifs.read((char*)&num_grouped, sizeof(uint32_t));

			assert(num_grouped < num_factors);

			for (uint i = 0; i < num_grouped; i++)
			{
				uint factor_index;
				ifs.read((char*)&factor_index, sizeof(uint32_t));

				CompIncreasingArrayP group_offsets = CompIncreasingArrayP(new CompIncreasingArray(ifs));
				MyPackedArrayP group_values = MyPackedArrayP(new MyPackedArray(ifs));

				groupings[factor_index] = morpho_groupingP(new morpho_grouping(group_values, group_offsets));
			}


			for (uint i = 0; i < num_factors; i++)
			{
				uint bpv;
				ifs.read((char*)&bpv, sizeof(uint32_t));
				assert(bpv < 64);
				bits_per_value.push_back(bpv);
			}
	
			for (uint i = 0; i < num_factors - 1; i++)
			{
				uint bpch;
				ifs.read((char*)&bpch, sizeof(uint32_t));
				assert(bpch < 64);
				bits_per_children.push_back(bpch);				
			}

			formOffsets = CompIncreasingArrayP(new CompIncreasingArray(ifs));
			morphoData = MyBitArrayP(new MyBitArray(ifs));
		}
		
	};


}
#endif //MORPHOLOGY_HPP_
