/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "MorphologyMapper.hpp"
#include "GlobalVars.hpp"
#include "StringMapper.hpp"
#include "MyPackedArray.hpp"
#include "CompIncreasingArray.hpp"
#include "MyMappedDoubleArray.hpp"

namespace ngramchecker {

		WordFactors::WordFactors(uint32_t _formID, uint32_t _lemmaID, uint32_t _tagID, double lemma_prob, double tag_prob)
		{
			factors[0] = _formID;
			factors[1] = _lemmaID;
			factors[2] = _tagID;

			emmision_probs[0] = 0;
			emmision_probs[1] = lemma_prob;
			emmision_probs[2] = tag_prob;
		}

		string WordFactors::ToString()
		{
			stringstream strs;
			strs << "Word = " << GlobalVars::stringMapper->InvMapForm(factors[0]) << endl;
			strs << "--- Lemma = " << GlobalVars::stringMapper->InvMapLemma(factors[1]) << ", em_prob = " << emmision_probs[1] << endl;
			strs << "--- Tag = " << GlobalVars::stringMapper->InvMapTag(factors[2]) << ", em_prob = " << emmision_probs[2] << endl;
			return strs.str();
		}

		uint32_t MorphologyMapper::NumForms()
		{
			return MA_offsets->GetSize();
		}

		uint32_t MorphologyMapper::NumLemmas()
		{
			return MA_lemmas->GetSize();
		}

		uint32_t MorphologyMapper::GetSentenceStartID()
		{
			return universal_sentence_start;
		}

		uint32_t MorphologyMapper::GetSentenceEndID()
		{
			return universal_sentence_end;
		}

		uint32_t MorphologyMapper::GetNumberTokenID()
		{
			return universal_number;
		}

		uint32_t MorphologyMapper::GetNameTokenID()
		{
			return universal_name;
		}

		uint32_t MorphologyMapper::GetUnkTokenID()
		{
			return universal_unk;
		}

		uint32_t MorphologyMapper::GetFirstNonSpecialID()
		{
			return universal_first_non_special;
		}

		void MorphologyMapper::createBinaryFiles()
		{
			ifstream istr;
			string s;

			ofstream ofs;
			ofs.open( MyConstants::filename_morphology_bin.c_str(), ios::binary | ios::out );
			assert(ofs.is_open());

			uint32_t universal_sentence_start = GlobalVars::stringMapper->MapForm("<s>");
			uint32_t universal_sentence_end = GlobalVars::stringMapper->MapForm("</s>");
			uint32_t universal_number = GlobalVars::stringMapper->MapForm("<number>");
			uint32_t universal_name = GlobalVars::stringMapper->MapForm("<name>");
			uint32_t universal_unk = GlobalVars::stringMapper->MapForm("<unk>");

			uint32_t universal_first_non_special = GlobalVars::stringMapper->MapForm("`") + 1;
			uint32_t universal_punctuation_start = GlobalVars::stringMapper->MapForm(".");

			ofs.write((char*)&universal_sentence_start, sizeof(uint32_t));
			ofs.write((char*)&universal_sentence_end, sizeof(uint32_t));
			ofs.write((char*)&universal_number, sizeof(uint32_t));
			ofs.write((char*)&universal_name, sizeof(uint32_t));
			ofs.write((char*)&universal_unk, sizeof(uint32_t));

			ofs.write((char*)&universal_first_non_special, sizeof(uint32_t));
			ofs.write((char*)&universal_punctuation_start, sizeof(uint32_t));

			istr.open( (MyConstants::project_home + "/data/morphology/text/tagGroups.out").c_str() );
			assert(istr.is_open());

			vector< vector<uint32_t> > tagGroups;
			while (std::getline(istr, s))
			{
				vector<string> vsTags;
				boost::split(vsTags, s, boost::is_any_of(";"));

				vector<uint32_t> tag_vec;
				for (uint32_t i = 0; i < vsTags.size(); i++)
				{
					tag_vec.push_back(boost::lexical_cast<uint32_t>(vsTags[i]));
				}

				tagGroups.push_back(tag_vec);
			}

			istr.close();

			uint32_t tag_groups_size = tagGroups.size();
			ofs.write((char*)&tag_groups_size, sizeof(uint32_t));

			for (uint32_t i = 0; i < tagGroups.size(); i++)
			{
				uint32_t curr_tag_group_size = tagGroups[i].size();
				ofs.write((char*)&curr_tag_group_size, sizeof(uint32_t));

				for (uint32_t j = 0; j < tagGroups[i].size(); j++)
				{
					uint32_t val = tagGroups[i][j];
					ofs.write((char*)&val, sizeof(uint32_t));
				}

			}


			istr.open((MyConstants::project_home + "/data/morphology/text/morphology.out").c_str());

			assert(istr.is_open());

			vector<uint32_t> ma_offsets;
			vector<uint32_t> ma_lemmas;
			vector<uint32_t> ma_tags;

			uint32_t actual_offset = 0;

			while (std::getline(istr, s))
			{
				ma_offsets.push_back(actual_offset);

				vector<string> parts;

				boost::split(parts, s, boost::is_any_of(" -"));

				//cerr_vector_string(parts);
				//cerr << endl;

				if (! (parts.size() > 2 && (parts.size() % 2 == 1)))
				{
					cerr << "corrupted line: " << s << endl;
				}
				assert(parts.size() > 2 && (parts.size() % 2 == 1));

				for (uint32_t i = 1; i < parts.size(); i += 2)
				{
					uint32_t lemmaID = boost::lexical_cast<uint32_t>(parts[i]);

					uint32_t tagGroupID = boost::lexical_cast<uint32_t>(parts[i + 1]);

					ma_lemmas.push_back(lemmaID);
					ma_tags.push_back(tagGroupID);

					actual_offset++;
				}
			}

			MyPackedArray MA_lemmas = MyPackedArray(ma_lemmas);
			MyPackedArray MA_tags = MyPackedArray(ma_tags);
			CompIncreasingArray MA_offsets = CompIncreasingArray(ma_offsets, MA_lemmas.GetSize() - 1);

			MA_offsets.WriteToStream(ofs);
			MA_lemmas.WriteToStream(ofs);
			MA_tags.WriteToStream(ofs);

			ofs.close();

			cout << "OK!!!";
		}


		MorphologyMapper::MorphologyMapper(bool init_emmisions, uint32_t cash_size): morphology_cash(cash_size)
		{
			ifstream ifs;
			ifs.open((MyConstants::filename_morphology_bin).c_str(), ios::binary);
			assert(ifs.is_open());

			ifs.read((char*)&universal_sentence_start, sizeof(uint32_t));
			ifs.read((char*)&universal_sentence_end, sizeof(uint32_t));
			ifs.read((char*)&universal_number, sizeof(uint32_t));
			ifs.read((char*)&universal_name, sizeof(uint32_t));
			ifs.read((char*)&universal_unk, sizeof(uint32_t));

			ifs.read((char*)&universal_first_non_special, sizeof(uint32_t));
			ifs.read((char*)&universal_punctuation_start, sizeof(uint32_t));

			MA_tagGroupsDefs = vector< vector<uint32_t> >();
			uint32_t tag_groups_size;

			ifs.read((char*)&tag_groups_size, sizeof(uint32_t));

			for (uint32_t i = 0; i < tag_groups_size; i++)
			{
				uint32_t curr_tag_group_size;
				ifs.read((char*)&curr_tag_group_size, sizeof(uint32_t));
				MA_tagGroupsDefs.push_back(vector<uint32_t>());

				for (uint32_t j = 0; j < curr_tag_group_size; j++)
				{
					uint32_t val;
					ifs.read((char*)&val, sizeof(uint32_t));
					MA_tagGroupsDefs[i].push_back(val);
				}

			}

			MA_offsets = CompIncreasingArrayP(new CompIncreasingArray(ifs));
			MA_lemmas = MyPackedArrayP(new MyPackedArray(ifs));
			MA_tagGroups = MyPackedArrayP(new MyPackedArray(ifs));


			ifs.close();

			if (init_emmisions)
			{
				ifs.open(MyConstants::filename_emmisions_bin.c_str(), ios::in | ios::binary);

				EM_lemma = MyMappedDoubleArrayP(new MyMappedDoubleArray(ifs));

				EM_tag = MyMappedDoubleArrayP(new MyMappedDoubleArray(ifs));
				EM_tag_offsets = CompIncreasingArrayP(new CompIncreasingArray(ifs));

				ifs.close();
			}
		}


		vectorP_WordFactors MorphologyMapper::GetOutputArcsPerWord(uint32_t formID, bool use_tag)
		{
			vectorP_WordFactors ret(new vector<WordFactors>());

			/*if (use_lemma == false && use_tag == false)
			{
				ret->push_back(WordFactors(formID, 0, 0, 0.0, 0.0));
				return ret;
			}*/

			pair<uint32_t, bool> cash_key = pair<uint32_t, bool>(formID, use_tag);
			if (morphology_cash.IsCashed(cash_key))
			{
				ret = morphology_cash.GetCashedValue(cash_key);
				return ret;
			}

			CompIA_First_Last_IndexPair index_pair = MA_offsets->GetFirstLastIndexPair(formID);

			assert(index_pair != NULL);

			uint32_t tag_em_offset;
			if (use_tag)
				tag_em_offset = EM_tag_offsets->GetValueAt(formID);
			else
				tag_em_offset = 0; //dummy value, in this case value never used!

			uint32_t offset_shift = 0;

			for (uint32_t i = index_pair->first; i <= index_pair->second; i++)
			{
				uint32_t lemmaID = MA_lemmas->GetValueAt(i);
				double lemma_em_prob = EM_lemma->GetValueAt(i);

				if (use_tag)
				{
					uint32_t tagGroupID = MA_tagGroups->GetValueAt(i);
					uint32_t tag_group_size = MA_tagGroupsDefs[tagGroupID].size();

					for (uint32_t j = 0; j < tag_group_size; j++)
					{
						double tag_em_prob = EM_tag->GetValueAt(tag_em_offset + offset_shift);
						offset_shift++;

						uint32_t tagID = MA_tagGroupsDefs[tagGroupID][j];
						ret->push_back(WordFactors(formID, lemmaID, tagID, lemma_em_prob, tag_em_prob));
					}
				}
				else
				{
					ret->push_back(WordFactors(formID, lemmaID, 0, lemma_em_prob, 0.0));
				}
			}

			morphology_cash.StoreValueForKey(cash_key, ret);

			return ret;
		}

		vectorP_WordFactors MorphologyMapper::GetOutputArcsPerWordNoEmmisions(uint32_t formID, bool use_tag)
		{
			vectorP_WordFactors ret(new vector<WordFactors>());

			/*if (use_lemma == false && use_tag == false)
			{
				ret->push_back(WordFactors(formID, 0, 0, 0.0, 0.0));
				return ret;
			}*/

			CompIA_First_Last_IndexPair index_pair = MA_offsets->GetFirstLastIndexPair(formID);

			if (index_pair == NULL)
			{
				cerr << "empty index pair! formID = " << formID << ", form_str = " << GlobalVars::stringMapper->InvMapForm(formID) << endl;
				cerr << "NumForms = " << MA_offsets->GetSize() << endl;
			}

			assert(index_pair != NULL);


			for (uint32_t i = index_pair->first; i <= index_pair->second; i++)
			{
				uint32_t lemmaID = MA_lemmas->GetValueAt(i);

				if (use_tag)
				{
					uint32_t tagGroupID = MA_tagGroups->GetValueAt(i);
					uint32_t tag_group_size = MA_tagGroupsDefs[tagGroupID].size();

					for (uint32_t j = 0; j < tag_group_size; j++)
					{
						uint32_t tagID = MA_tagGroupsDefs[tagGroupID][j];
						ret->push_back(WordFactors(formID, lemmaID, tagID, 0.0, 0.0));
					}
				}
				else
				{
					ret->push_back(WordFactors(formID, lemmaID, 0, 0.0, 0.0));
				}
			}


			return ret;
		}


		void MorphologyMapper::MakeEmmisionSmoothingTextFile()
		{
			ofstream ostr;
			ostr.open((MyConstants::project_home + "/data/lm/word_emmision/smooth_lemma.counts").c_str());
			assert(ostr.is_open());

			assert(GlobalVars::stringMapper->NumForms() == this->NumForms());

			for (uint32_t i = 0; i < GlobalVars::stringMapper->NumForms(); i++)
			{
				vectorP_WordFactors vwf = this->GetOutputArcsPerWordNoEmmisions(i, false);

				string form = GlobalVars::stringMapper->InvMapForm(i);

				for (uint32_t j = 0; j < vwf->size(); j++)
				{
					string lemma = GlobalVars::stringMapper->InvMapLemma(vwf->operator [](j).factors[1]);
					ostr << form << " " << lemma << endl;
				}
			}

			ostr.close();

			ostr.open((MyConstants::project_home + "/data/lm/word_emmision/smooth_tag.counts").c_str());
			assert(ostr.is_open());

			for (uint32_t i = 0; i < GlobalVars::stringMapper->NumForms(); i++)
			{
				vectorP_WordFactors vwf = this->GetOutputArcsPerWordNoEmmisions(i, true);

				string form = GlobalVars::stringMapper->InvMapForm(i);

				for (uint32_t j = 0; j < vwf->size(); j++)
				{
					string tag = GlobalVars::stringMapper->InvMapTag(vwf->operator [](j).factors[2]);
					ostr << form << " " << tag << endl;
				}

			}

			ostr.close();

			cerr << "OK!";
		}

		void MorphologyMapper::MakeWordEmissionBinFilesFromTextFiles(uint32_t bits_per_lemma, uint32_t bits_per_tag)
		{
			ifstream istr;
			istr.open((MyConstants::project_home + "/data/lm/word_emmision/lemma.probs").c_str());
			assert(istr.is_open());

			string s;

			vector<double> em_lemma;

			for (uint32_t formID = 0; formID < MA_offsets->GetSize(); formID++)
			{

				CompIA_First_Last_IndexPair index_pair = MA_offsets->GetFirstLastIndexPair(formID);

				if (index_pair == NULL)
				{
					cerr << "empty index pair! formID = " << formID << ", form_str = " << GlobalVars::stringMapper->InvMapForm(formID) << endl;
					cerr << "NumForms = " << MA_offsets->GetSize() << endl;
				}
				assert(index_pair != NULL);

				string form_str = GlobalVars::stringMapper->InvMapForm(formID);

				for (uint32_t j = index_pair->first; j <= index_pair->second; j++)
				{
					uint32_t lemmaID = MA_lemmas->GetValueAt(j);
					string lemma_str = GlobalVars::stringMapper->InvMapLemma(lemmaID);

					std::getline(istr, s);

					vector<string> parts;
					boost::split(parts, s , boost::is_any_of(" "));

					assert(parts[0] == form_str);
					assert(parts[1] == lemma_str);
					assert(parts.size() == 3);

					double em_prob = boost::lexical_cast<double>(parts[2]);
					em_lemma.push_back(em_prob);
				}

			}

			MyMappedDoubleArray mmda = MyMappedDoubleArray(em_lemma, bits_per_lemma);

			ofstream ostr;
			ostr.open((MyConstants::filename_emmisions_bin).c_str(), ios::out | ios::binary);
			assert(ostr.is_open());

			mmda.WriteToStream(ostr);

			istr.close();

			istr.open((MyConstants::project_home + "/data/lm/word_emmision/tag.probs").c_str());
			assert(istr.is_open());

			map<string, double> prob_map;

			uint32_t pom_counter = 0;
			while (std::getline(istr, s))
			{
				vector<string> parts;
				boost::split(parts, s, boost::is_any_of(" "));

				double single_prob = boost::lexical_cast<double>(parts[2]);

				//if (pom_counter % 1000 == 0)
				//	cerr << "single_prob = " << single_prob << endl;

				prob_map[ parts[0] + " " + parts[1] ] = single_prob;

				pom_counter++;
			}

			istr.close();

			uint32_t pom_pointer = 0;
			vector<uint32_t> tag_offsets;
			vector<double> em_tag;

			for (uint32_t formID = 0; formID < GlobalVars::stringMapper->NumForms(); formID++)
			{
				tag_offsets.push_back(pom_pointer);
				string form = GlobalVars::stringMapper->InvMapForm(formID);
				vectorP_WordFactors vwf = this->GetOutputArcsPerWordNoEmmisions(formID, true);

				assert(vwf->size() > 0);

				for (uint32_t j = 0; j < vwf->size(); j++)
				{
					uint32_t tagID = vwf->operator [](j).factors[2];
					string tag = GlobalVars::stringMapper->InvMapTag(tagID);

					assert(prob_map.find(form + " " + tag) != prob_map.end());

					em_tag.push_back(prob_map[form + " " + tag]);
					pom_pointer++;
				}
			}

			MyMappedDoubleArray mm_em_probs_tags = MyMappedDoubleArray(em_tag, bits_per_tag);
			mm_em_probs_tags.WriteToStream(ostr);

			CompIncreasingArray myarray_tag_offsets = CompIncreasingArray(tag_offsets, mm_em_probs_tags.GetSize() - 1);
			myarray_tag_offsets.WriteToStream(ostr);


			ostr.close();

			cerr << "MakeWordEmissionBinFiles -> OK!" << endl;
		}


		uint32_t MorphologyMapper::PunctuationUniversalIndex_char(char character)
		{
			return GlobalVars::stringMapper->PunctuationIndex_char(character) + universal_punctuation_start;
		}

		bool MorphologyMapper::IsNormalWord(uint32_t form_id)
		{
			return form_id >= universal_first_non_special;
		}

		void MorphologyMapper::TestIt(string outfilename)
		{

			ofstream ofs;
			ofs.open(outfilename.c_str());
			assert(ofs.is_open());


			for (uint32_t formID = 0; formID < NumForms(); formID++)
			{
				string lastLemma = "";
				vectorP_WordFactors wordMorpho = GetOutputArcsPerWordNoEmmisions(formID, true);
				ofs << "<f>" << GlobalVars::stringMapper->InvMapForm(wordMorpho->operator [](0).factors[0]);

				for (uint32_t j = 0; j < wordMorpho->size(); j++)
				{
					string lemma = GlobalVars::stringMapper->InvMapLemma(wordMorpho->operator [](j).factors[1]);
					if (lemma != lastLemma)
					{
						lastLemma = lemma;
						ofs << "<l>" << lemma;
					}

					string tag = GlobalVars::stringMapper->InvMapTag(wordMorpho->operator [](j).factors[2]);
					ofs << "<t>" << tag;
				}

				ofs << endl;
			}

			ofs.close();
			cerr << "Test finished!";
		}

}