/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */
#include "Spellchecker.hpp"
#include "GlobalVars.hpp"
#include "StringMapper.hpp"
#include "MorphologyMapper.hpp"
#include "LemmaSigns.hpp"
#include "TextCheckingResult.hpp"
#include "Token.hpp"
#include "Splitter.hpp"
//#include "StagePosibility.hpp"
#include "TransitionCostComputation.hpp"
#include "DecoderBase.hpp"

namespace ngramchecker {

	map<uint32_t, vector<StagePosibilityP> > Spellchecker::MakeSuggestionList(vector<StagePosibilityP> &decoded_pos, StagePosibilitiesType stage_posibilities) {

			assert(stage_posibilities->size() == decoded_pos.size());

			//cerr << "Make suggestion started..." << endl;
			uint32_t viterbi_order = decoder->GetViterbiOrder();

			map<uint32_t, vector<StagePosibilityP> > suggestions;


			for (uint32_t i = viterbi_order - 1; i < decoded_pos.size() - 1; i++)
			{
				//cerr << "i = " << i << endl;

				if (decoded_pos[i]->IsOriginal())
				{
					if (decoded_pos[i]->IsUnknown())
					{
						suggestions[i] = vector<StagePosibilityP>();
					}

					continue;
				}


				//assert (i < stage_posibilities->size());
				//assert (stage_posibilities != NULL);
				vector<StagePosibilityP> st_pos = stage_posibilities->operator [](i);

				vector<pair<StagePosibilityP, double> > cost_vector;

				suggestions[i] = vector<StagePosibilityP>();

				StagePosibilityP decoded_pos_at_i_backup = decoded_pos[i];

				for (uint32_t j = 0; j < st_pos.size(); j++)
				{
					//cerr << "j = " << j << endl;
					StagePosibilityP sp = st_pos[j];
					double cost = sp->EmmisionProbability();

					decoded_pos[i] = sp;

					for (uint32_t k = 0; k < viterbi_order; k++)
					{
						if (i + k >= decoded_pos.size() )
							break;

						cost += transition_cost_computation->ComputeTransitionCostSPSequence(decoded_pos, i - viterbi_order + 1 + k, i + k);
					}

					cost_vector.push_back(pair<StagePosibilityP, double>(sp, cost));

				}

				decoded_pos[i] = decoded_pos_at_i_backup;

				std::sort(cost_vector.begin(), cost_vector.end(), Pair_StagePosibilityP_double_comparer());
				vector<pair<StagePosibilityP, double> > new_cost_vector;
				set<StagePosibilityP, StagePosibility_Form_comparer> word_ids_in_list;

				for (uint32_t j = 0; j < cost_vector.size(); j++)
				{
					if (cost_vector[j].first->IsOriginal())
						break;

					if (word_ids_in_list.find(cost_vector[j].first) == word_ids_in_list.end())
					{
						word_ids_in_list.insert(cost_vector[j].first);
						new_cost_vector.push_back(cost_vector[j]);
					}
				}


				for (vector<pair<StagePosibilityP, double> >::iterator it = new_cost_vector.begin(); it != new_cost_vector.end(); it++)
				{
					suggestions[i].push_back(it->first);
				}


			}

			//cerr << "... make suggestion finished!" << endl;
			return suggestions;

		}

		vector<TextCheckingResultP> Spellchecker::GetCheckingResults(string sentence)
		{
			vector<TextCheckingResultP> results;
			vector<string> tokens_str;
			vector<string> tok_delims;
			Splitter::SplitSentenceToTokens(sentence, tokens_str, tok_delims);

			vector<Token> tokens;

			for (uint32_t i = 0; i < tokens_str.size(); i++)
				tokens.push_back(Token(tokens_str[i]));

			vector<pair<uint32_t,uint32_t> > token_ranges;
			uint32_t curr_pos = 0;

			for (uint32_t i = 0; i < tokens_str.size(); i++)
			{
				token_ranges.push_back(pair<uint32_t, uint32_t>(curr_pos, tokens_str[i].length()) );
				curr_pos += tokens_str[i].length();
				if (i < tok_delims.size())
					curr_pos += tok_delims[i].length();
			}


			vector<StagePosibilityP> spv;
			StagePosibilitiesType stage_posibs;
			decoder->DecodeTokenizedSentence_ReturnStagePosibilities(tokens, spv, stage_posibs);

			uint32_t decoder_order = decoder->GetViterbiOrder();

			map<uint32_t, vector<StagePosibilityP> > sugg = MakeSuggestionList(spv, stage_posibs);

			for (map<uint32_t, vector<StagePosibilityP> >::iterator it = sugg.begin(); it != sugg.end(); it++)
			{
				assert(it->first >= decoder_order - 1);
				uint32_t tok_index = it->first - (decoder_order - 1);
				vector<StagePosibilityP> stage_pos = it->second;

				vector<string> suggestions;

				for (uint32_t i = 0; i < stage_pos.size(); i++)
				{
					suggestions.push_back(stage_pos[i]->ToString());
				}

				//cerr << tokens[tok_index].GetWordStringUTF8() << "sugg size: " << suggestions.size() << endl;

				if (suggestions.size() > 0 &&
					GlobalVars::stringMapper->CaseInsensitiveIsEqual(tokens[tok_index].GetWordStringISO(), suggestions[0]))
				{
					string sugg = suggestions[0];
					suggestions.clear();
					suggestions.push_back(sugg);
					results.push_back(TextCheckingResultP(new GrammarCheckingResult(token_ranges[tok_index].first, token_ranges[tok_index].second, suggestions, "Error in capitalization? Did you mean " + suggestions[0] + "?", tokens[tok_index].GetWordStringUTF8())));
				}
				else if (tokens[tok_index].ID == GlobalVars::morphologyMapper->GetNameTokenID() ||
					tokens[tok_index].ID == GlobalVars::morphologyMapper->GetUnkTokenID())
				{
						results.push_back(TextCheckingResultP(new SpellingCheckingResult(token_ranges[tok_index].first, token_ranges[tok_index].second, suggestions, tokens[tok_index].GetWordStringUTF8())));
				}
				else if (suggestions.size() > 0) //It's not intuitive that size can be zero! However it's possible with agressive prunning during the decoding!
				{
					results.push_back(TextCheckingResultP(new GrammarCheckingResult(token_ranges[tok_index].first, token_ranges[tok_index].second, suggestions, "Did you mean " + suggestions[0] + "?", tokens[tok_index].GetWordStringUTF8())));
				}
			}

			for (uint32_t i = 0; i < tokens.size(); i++)
			{
				if (GlobalVars::lemmaSigns->IsFormOkey(tokens[i].ID) == false)
				{
					vector<string> emptySuggestions;
					results.push_back(boost::shared_ptr<TextCheckingResult>(new GrammarCheckingResult(token_ranges[i].first, token_ranges[i].second, emptySuggestions, "Word is inappropriate (vulgar, too expresive, archaic etc...)", tokens[i].GetWordStringUTF8() )));
				}
			}


			return results;
		}

		string Spellchecker::CheckSentence(const string sentence)
		{
			return decoder->DecodeSentence(sentence);
		}

		string Spellchecker::DecodeEvaluation(const string sentence, uint32_t num_sugg_to_output)
		{
				stringstream ret;

				uint32_t decoder_order = decoder->GetViterbiOrder();

				vector<string> tokens_str;
				vector<string> tok_delims;
				Splitter::SplitSentenceToTokens(sentence, tokens_str, tok_delims);

				vector<Token> tokens;

				for (uint32_t i = 0; i < tokens_str.size(); i++)
					tokens.push_back(Token(tokens_str[i]));

				vector<StagePosibilityP> spv;
				StagePosibilitiesType stage_posibs;

				decoder->DecodeTokenizedSentence_ReturnStagePosibilities(tokens, spv, stage_posibs);

				map<uint32_t, vector<StagePosibilityP> > sugg = MakeSuggestionList(spv, stage_posibs);

				for (uint32_t i = 0; i < tokens.size(); i++)
				{
					if (i > 0) ret << " ";

					map<uint32_t, vector<StagePosibilityP> >::iterator fit = sugg.find(i + decoder_order - 1);

					if (fit == sugg.end())
					{
						if (spv[i + decoder_order - 1]->IsUnknown())
						{
							//cerr << "spv->IsUnknown: " << tokens[i].GetWordStringISO() << endl;
							ret << tokens[i].GetWordStringISO() << "(spelling[])";
						}
						else
							ret << tokens[i].GetWordStringISO();
					}
					else
					{
						ret << tokens[i].GetWordStringISO();

						if (tokens[i].IsUnknown())
							ret << "(spelling[";
						else
							ret << "(grammar[";

						vector<StagePosibilityP> vec_sp = fit->second;

						//assert(vec_sp.size() > 0);

						for (uint32_t g = 0; g < min(vec_sp.size(), (size_t)num_sugg_to_output); g++)
						{
							/*if (g == 0)
							{
								//cerr << "looking for " << spv[i + decoder_order - 2]->ToString() << " --> " << vec_sp[g]->ToString() <<
								//		" or " << vec_sp[g]->ToString() << " --> " << spv[i + decoder_order]->ToString() << endl;
								uint32_t next_form = spv[i + decoder_order]->FormIdentifier();
								uint32_t prev_form = spv[i + decoder_order - 2]->FormIdentifier();
								uint32_t act_form = vec_sp[g]->FormIdentifier();
								if (GlobalVars::lmServer->ExistFormBigram(act_form, prev_form) &&
									GlobalVars::lmServer->ExistFormBigram(next_form, act_form) &&
									next_form != GlobalVars::morphologyMapper->GetNameTokenID() &&
									next_form != GlobalVars::morphologyMapper->GetUnkTokenID() &&
									prev_form != GlobalVars::morphologyMapper->GetNameTokenID() &&
									prev_form != GlobalVars::morphologyMapper->GetUnkTokenID()
								   )
								{
									ret << "!";
								}
							}*/

							if (g > 0) ret << "|";
							ret << vec_sp[g]->ToString();
						}

						ret << "])";
					}
				}


			return ret.str();
		}
		
		string Spellchecker::command_line_mode(const string sentence, uint32_t num_sugg_to_output)
		{
				stringstream ret;

				uint32_t decoder_order = decoder->GetViterbiOrder();

				vector<string> tokens_str;
				vector<string> tok_delims;
				Splitter::SplitSentenceToTokens(sentence, tokens_str, tok_delims);

				vector<Token> tokens;

				for (uint32_t i = 0; i < tokens_str.size(); i++)
					tokens.push_back(Token(tokens_str[i]));

				vector<StagePosibilityP> spv;
				StagePosibilitiesType stage_posibs;

				decoder->DecodeTokenizedSentence_ReturnStagePosibilities(tokens, spv, stage_posibs);

				map<uint32_t, vector<StagePosibilityP> > sugg = MakeSuggestionList(spv, stage_posibs);

				for (uint32_t i = 0; i < tokens.size(); i++)
				{
					if (i > 0) ret << tok_delims[i - 1];

					map<uint32_t, vector<StagePosibilityP> >::iterator fit = sugg.find(i + decoder_order - 1);

					if (fit == sugg.end())
					{
						if (spv[i + decoder_order - 1]->IsUnknown())
						{
							//cerr << "spv->IsUnknown: " << tokens[i].GetWordStringISO() << endl;
							ret << "<spelling original=\"" << tokens[i].GetWordStringISO() << "\" suggestions=\"\"/>";
						}
						else
							ret << tokens[i].GetWordStringISO();
					}
					else
					{

						if (tokens[i].IsUnknown())
							ret << "<spelling original=\"";
						else
							ret << "<grammar original=\"";

						ret << tokens[i].GetWordStringISO();
						
						ret << "\" suggestions=\"";

						vector<StagePosibilityP> vec_sp = fit->second;

						//assert(vec_sp.size() > 0);

						for (uint32_t g = 0; g < min(vec_sp.size(), (size_t)num_sugg_to_output); g++)
						{
							/*if (g == 0)
							{
								//cerr << "looking for " << spv[i + decoder_order - 2]->ToString() << " --> " << vec_sp[g]->ToString() <<
								//		" or " << vec_sp[g]->ToString() << " --> " << spv[i + decoder_order]->ToString() << endl;
								uint32_t next_form = spv[i + decoder_order]->FormIdentifier();
								uint32_t prev_form = spv[i + decoder_order - 2]->FormIdentifier();
								uint32_t act_form = vec_sp[g]->FormIdentifier();
								if (GlobalVars::lmServer->ExistFormBigram(act_form, prev_form) &&
									GlobalVars::lmServer->ExistFormBigram(next_form, act_form) &&
									next_form != GlobalVars::morphologyMapper->GetNameTokenID() &&
									next_form != GlobalVars::morphologyMapper->GetUnkTokenID() &&
									prev_form != GlobalVars::morphologyMapper->GetNameTokenID() &&
									prev_form != GlobalVars::morphologyMapper->GetUnkTokenID()
								   )
								{
									ret << "!";
								}
							}*/

							if (g > 0) ret << "|";
							ret << vec_sp[g]->ToString();
						}

						ret << "\"/>";
					}
				}


			return ret.str();
		}

		Spellchecker::Spellchecker(DecoderBaseP _decoder):
			decoder(_decoder), transition_cost_computation(_decoder->GetTransitionCostComputation())
		{}

}