/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#ifndef STAGEPOSIBILITY_HPP_
#define STAGEPOSIBILITY_HPP_

#include "StdAfx.h"

namespace ngramchecker {

	class StagePosibility {
	public:
		StagePosibility_type type;

		virtual double EmmisionProbability() = 0;
		virtual string ToString() = 0;
		virtual bool IsOriginal() = 0;
		virtual uint32_t UniqueIdentifier() = 0;
		virtual uint32_t FormIdentifier() = 0;
		virtual bool IsUnknown() = 0;

		StagePosibility(StagePosibility_type _type): type(_type) {}
	};

	SP_DEF(StagePosibility);


	struct StagePosibility_Identifying_comparer : less<StagePosibilityP> {
		bool operator()(const StagePosibilityP& sp1, const StagePosibilityP& sp2)
		{
			return sp1->UniqueIdentifier() < sp2->UniqueIdentifier();
		}
	};

	struct StagePosibility_Form_comparer : less<StagePosibilityP> {
		bool operator()(const StagePosibilityP& sp1, const StagePosibilityP& sp2)
		{
			return sp1->FormIdentifier() < sp2->FormIdentifier();
		}
	};


	struct StagePosibility_sort_cost : less<StagePosibilityP> {
		bool operator()(const StagePosibilityP& sp1, const StagePosibilityP& sp2)
		{
			return sp1->EmmisionProbability() < sp2->EmmisionProbability();
		}
	};

	class StagePosibility_FormsOnly : public StagePosibility {

	private:
		uint32_t id;
		string word;
		bool original;
		double emmision_prob;

	public:

		StagePosibility_FormsOnly(uint32_t _id, string _word, bool _original, double _emmision_prob):
			StagePosibility(FormsOnly), id(_id), word(_word), original(_original), emmision_prob(_emmision_prob)
		{}

		virtual double EmmisionProbability();

		virtual string ToString();

		virtual uint32_t UniqueIdentifier();

		virtual uint32_t FormIdentifier();

		virtual bool IsOriginal();

		virtual bool IsUnknown();

		static StagePosibilityP SentenceStartSP();

		static StagePosibilityP SentenceEndSP();

		uint32_t ID();
	};

	SP_DEF(StagePosibility_FormsOnly);

	class StagePosibility_MultiFactor : public StagePosibility
	{
		uint32_t formID;
		uint32_t lemmaID;
		uint32_t tagID;
		string word;
		bool original;
		double emmision_prob;
		size_t uniq_id;


		StagePosibility_MultiFactor(uint32_t _formID, uint32_t _lemmaID, uint32_t _tagID, string _word, bool _original, double _emmision_cost);

	public:

		static StagePosibilityP CreateSP_ZeroEmmisionCost(uint32_t formID, uint32_t lemmaID, uint32_t tagID, string word, bool original);

		static StagePosibilityP CreateSP_ErrorModelCost(uint32_t formID, uint32_t lemmaID, uint32_t tagID, string word, bool original, double error_model_cost);

		static StagePosibilityP CreateSP_ComplexEmCost(uint32_t formID, uint32_t lemmaID, uint32_t tagID, string word, bool original, double error_model_cost, double lemma_em_cost, double tag_em_cost, double lemma_weight, double tag_weight);

		virtual double EmmisionProbability();

		virtual string ToString();

		virtual uint32_t UniqueIdentifier();

		virtual uint32_t FormIdentifier();

		virtual bool IsOriginal();

		virtual bool IsUnknown();

		static StagePosibilityP SentenceStartSP();

		static StagePosibilityP SentenceEndSP();

		uint32_t FormID();

		uint32_t LemmaID();

		uint32_t TagID();

	};

	SP_DEF(StagePosibility_MultiFactor);

	class StagePosibility_MorphoAnal : public StagePosibility
	{
		uint32_t formID;
		uint32_t lemmaID;
		uint32_t tagID;

		double emmision_prob;
		size_t uniq_id;

	public:

		static const bool output_form = true;
		static const bool output_lemma = true;
		static const bool output_tag = true;

		StagePosibility_MorphoAnal(uint32_t _formID, uint32_t _lemmaID, uint32_t _tagID, double _lemma_emmision_cost, double _tag_emmision_cost, double _lemma_weight, double _tag_weight);

		virtual double EmmisionProbability();

		virtual string ToString();

		virtual uint32_t UniqueIdentifier();

		virtual uint32_t FormIdentifier();

		virtual bool IsOriginal();

		virtual bool IsUnknown();

		static StagePosibilityP SentenceStartSP();

		static StagePosibilityP SentenceEndSP();

		uint32_t FormID();

		uint32_t LemmaID();

		uint32_t TagID();
	};

	SP_DEF(StagePosibility_MorphoAnal);

	class StagePosibility_Letter : public StagePosibility
	{

	private:
		uint32_t letter_ID;
		double emmision_prob;
		string str_form;

	public:
		StagePosibility_Letter(uint32_t _letter, double _emmision_prob);

		virtual double EmmisionProbability();

		virtual string ToString();

		virtual bool IsOriginal();

		virtual uint32_t UniqueIdentifier();

		virtual uint32_t FormIdentifier();

		virtual bool IsUnknown();

		uint32_t LetterID();

		static StagePosibilityP SentenceStartSP();

		static StagePosibilityP SentenceEndSP();

	};

	SP_DEF(StagePosibility_Letter);

}


#endif /* STAGEPOSIBILITY_HPP_ */
