/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "StringMapper.hpp"
#include "GlobalVars.hpp"

namespace ngramchecker {

		void StringMapper::initFormMaps()
		{
			cerr << "Warning: Form map loaded!" << endl;
			form_map = boost::shared_ptr<map<string, uint32_t> >(new map<string, uint32_t>());
			form_inv_map = boost::shared_ptr<map<uint32_t, string> >(new map<uint32_t, string>());

			ifstream ifs;
			string s;
			ifs.open(MyConstants::filename_form_map.c_str());
			assert(ifs.is_open());
			uint32_t counter = 0;

			while (std::getline(ifs, s))
			{
				form_map->operator [](s) = counter;
				form_inv_map->operator [](counter) = s;
				counter++;
			}

			num_forms = counter;
		}

		void StringMapper::initLemmaMaps()
		{
			cerr << "Warning: Lemma map loaded!" << endl;
			lemma_map = boost::shared_ptr<map<string, uint32_t> >(new map<string, uint32_t>());
			lemma_inv_map = boost::shared_ptr<map<uint32_t, string> >(new map<uint32_t, string>());

			ifstream ifs;
			string s;
			ifs.open(MyConstants::filename_lemma_map.c_str());
			assert(ifs.is_open());
			uint32_t counter = 0;

			while (std::getline(ifs, s))
			{
				lemma_map->operator [](s) = counter;
				lemma_inv_map->operator [](counter) = s;
				counter++;
			}

			num_lemmas = counter;
		}

		void StringMapper::initTagMaps()
		{
			cerr << "Warning: Tag map loaded!" << endl;

			tag_map = boost::shared_ptr<map<string, uint32_t> >(new map<string, uint32_t>());
			tag_inv_map = boost::shared_ptr<map<uint32_t, string> >(new map<uint32_t, string>());

			ifstream ifs;
			string s;
			ifs.open(MyConstants::filename_tag_map.c_str());
			assert(ifs.is_open());
			uint32_t counter = 0;

			while (std::getline(ifs, s))
			{
				tag_map->operator [](s) = counter;
				tag_inv_map->operator [](counter) = s;
				counter++;
			}

			num_tags = counter;

		}

			bool StringMapper::SymbolIsPunctuation(uint32_t character)
			{
				return character >= punctuation_start_index && character <= punctuation_end_index;
			}

			bool StringMapper::SymbolIsLetter(uint32_t character)
			{
				return character < punctuation_start_index;
			}

			uint32_t StringMapper::GetLowerCaseForUpperCase_symbol(uint32_t letterIndex)
			{
				assert(letterIndex >= capitals_start_index && letterIndex < punctuation_start_index);

				return letterIndex - capitals_start_index;
			}

			uint32_t StringMapper::GetUpperCaseForLowerCase_symbol(uint32_t letterIndex)
			{
				assert(letterIndex < capitals_start_index);

				return letterIndex + capitals_start_index;
			}

			bool StringMapper::LetterIsDiacritic(uint32_t letter)
			{
				return diac_inv_map.find(letter) != diac_inv_map.end();
			}

			uint32_t StringMapper::GetBasicLetterForDiacriticLetter(uint32_t diacritic_letter)
			{
				assert(LetterIsDiacritic(diacritic_letter));

				return diac_inv_map[diacritic_letter];
			}

			bool StringMapper::LetterHasDiacriticVariants(uint32_t letter)
			{
				return diac_map.find(letter) != diac_map.end();
			}

			StringMapper::StringMapper()
			{
				ifstream ifs;
				ifs.open(MyConstants::filename_chars_list.c_str() );
				assert(ifs.is_open());
				string s;

				std::getline(ifs, s);


				charmap = map<char, uint32_t>();
				invcharmap = map<uint32_t, char>();

				for (uint32_t i = 0; i < s.length(); i++)
				{
					charmap[ s[i] ] = i;
					invcharmap[ i ] = s[i];
				}

				punctuation_start_index =  s.find('.');
				punctuation_end_index = s.find('0') - 1;

				number_start_index = s.find('0');
				number_end_index = s.find('9');


				std::getline(ifs, s);
				
				vocal_set = set<char>();
				
				for (uint32_t i = 0; i < s.length(); i++)
					vocal_set.insert(s[i]);
				

				ifs.close();
				
				capitals_start_index = charmap['A'];

				num_forms = 0;
				num_lemmas = 0;
				num_tags = 0;

				ifstream ifs_utf8;
				ifs_utf8.open( MyConstants::filename_utf8_list.c_str() );
				assert(ifs_utf8.is_open());

				ifstream ifs_iso;
				ifs_iso.open( MyConstants::filename_iso_list.c_str() );
				assert(ifs_iso.is_open());

				string iso_s;
				string utf8_s;

				while (std::getline(ifs_utf8, utf8_s))
				{
					std::getline(ifs_iso, iso_s);

					utf8_iso_map[pair<char, char>(utf8_s[0], utf8_s[1])] = iso_s[0];
					iso_utf8_map[iso_s[0]] = pair<char, char>(utf8_s[0], utf8_s[1]);

					for (uint32_t i = 0; i < utf8_s.length(); i++)
					{
						//cerr << (int)(utf8_s[i]) << " ";
					}
					//cerr << "=~ " << iso_s[0] << endl;
				}


				ifstream ifs_diac;
				ifs_diac.open(MyConstants::filename_error_model_diacritics.c_str());
				assert(ifs_diac.is_open());



				while (std::getline(ifs_diac, s))
				{
					vector<string> parts;

					boost::split(parts, s, boost::is_any_of(" "));

					string letters = parts[0];

					vector<uint32_t> dl_vec;
					for (uint32_t i = 1; i < letters.length(); i++)
					{
						dl_vec.push_back(MapChar(letters[i]));
						diac_inv_map[MapChar(letters[i])] = MapChar(letters[0]);
						//substitutions[pair<uint32_t, uint32_t>(MapChar(letters[0]), MapChar(letters[i]))] = boost::lexical_cast<double>(parts[i]);
					}

					diac_map[MapChar(letters[0])] = dl_vec;
				}

				ifs_diac.close();

				//cerr << "String mapper instance constructed!\n";
			}


		bool StringMapper::CharIsNumber(char ch)
		{
			return ch >= '0' && ch <= '9';
		}

		bool StringMapper::WordIsNumber(const string &word)
		{
			for (uint32_t i = 0; i < word.size(); i++)
			{
				if (!CharIsNumber(word[i]))
					return false;
			}

			return true;
		}

		bool StringMapper::CharIsPunctuation(char ch)
		{
			if (ch >= '0' && ch <= '9')
				return false;
			map<char, uint32_t>::iterator it = charmap.find(ch);
			if (it != charmap.end())
			{
				return SymbolIsPunctuation(it->second);
			}
			else
			{
				return false;
			}
		}
		
		bool StringMapper::CharIsVocal(char ch)
		{
			return (vocal_set.find(ch) != vocal_set.end());
		}

		bool StringMapper::CharIsKnown(char ch)
		{
			return charmap.find(ch) != charmap.end();
		}

		bool StringMapper::CharIsLetter(char ch)
		{
			map<char, uint32_t>::iterator it = charmap.find(ch);
			if (it != charmap.end())
			{
				return SymbolIsLetter(it->second);
			}
			else
			{
				return false;
			}
		}

		bool StringMapper::CharIsLowerCase(char ch)
		{
			map<char, uint32_t>::iterator it = charmap.find(ch);
			if (it != charmap.end())
			{
				return it->second < capitals_start_index;
			}
			else
			{
				return false;
			}
		}

		bool StringMapper::CharIsUpperCase(char ch)
		{
			map<char, uint32_t>::iterator it = charmap.find(ch);
			if (it != charmap.end())
			{
				return it->second >= capitals_start_index && it->second < punctuation_start_index;
			}
			else
			{
				return false;
			}
		}

		uint32_t StringMapper::NumForms()
		{
			if (num_forms == 0)
			{
				initFormMaps();
			}

			assert(num_forms != 0);
			return num_forms;
		}

		uint32_t StringMapper::NumLemmas()
		{
			if (num_lemmas == 0)
			{
				initLemmaMaps();
			}

			assert (num_lemmas != 0);
			return num_lemmas;
		}

		uint32_t StringMapper::NumTags()
		{
			if (num_tags == 0)
			{
				initTagMaps();
			}

			assert(num_tags != 0);
			return num_tags;
		}

		string StringMapper::InvMapForm(uint32_t id)
		{
			if (form_inv_map == NULL)
			{
				initFormMaps();
			}

			if (form_inv_map->find(id) == form_inv_map->end())
			{
				return "!!!Unk!!!";
			}

			return form_inv_map->operator [](id);
		}

		string StringMapper::InvMapLemma(uint32_t id)
		{
			if (lemma_inv_map == NULL)
			{
				initLemmaMaps();
			}

			assert(lemma_inv_map->find(id) != lemma_inv_map->end());

			return lemma_inv_map->operator [](id);
		}

		string StringMapper::InvMapTag(uint32_t id)
		{
			if (tag_inv_map == NULL)
			{
				initTagMaps();
			}

			assert(tag_inv_map->find(id) != tag_inv_map->end());

			return tag_inv_map->operator [](id);
		}



		uint32_t StringMapper::MapForm(string form)
		{
			if (form_map == NULL)
			{
				initFormMaps();
			}

			assert(form_map->find(form) != form_map->end());

			return form_map->operator [](form);
		}

		bool StringMapper::ContainsLemma(string lemma)
		{
			return lemma_map->find(lemma) != lemma_map->end();
		}

		uint32_t StringMapper::MapLemma(string lemma)
		{
			if (lemma_map == NULL)
			{
				initLemmaMaps();
			}

			assert(lemma_map->find(lemma) != lemma_map->end());

			return lemma_map->operator [](lemma);
		}

		uint32_t StringMapper::MapTag(string tag)
		{
			if (tag_map == NULL)
			{
				initTagMaps();
			}

			assert(tag_map->find(tag) != tag_map->end());

			return tag_map->operator [](tag);
		}

		uint32_t StringMapper::MapChar(char ch)
		{

			if (charmap.find(ch) == charmap.end())
			{
				cerr << "!!! Unknown character: " << ch << endl;
			}

			assert(charmap.find(ch) != charmap.end());

			return charmap[ch];

		}


		uint32_t StringMapper::PunctuationIndex_char(char character)
		{
			uint32_t letterID = MapChar(character);
			assert(letterID >= punctuation_start_index && letterID < number_start_index);

			return letterID - punctuation_start_index;
		}


		char StringMapper::InvMapChar(uint32_t val)
		{
			assert(invcharmap.find(val) != invcharmap.end());
			return invcharmap[val];
		}


		char StringMapper::GetLowerCaseForUpperCase_char(char character)
		{
			uint32_t symbol = MapChar(character);
			uint32_t lc_symbol = GetLowerCaseForUpperCase_symbol(symbol);
			return InvMapChar(lc_symbol);
		}

		char StringMapper::GetUpperCaseForLowerCase_char(char character)
		{
			uint32_t symbol = MapChar(character);
			uint32_t uc_symbol = GetUpperCaseForLowerCase_symbol(symbol);
			return InvMapChar(uc_symbol);
		}

		string StringMapper::InvMap(vector<uint32_t> &vec)
		{
			char* c_str = new char[vec.size() + 1];

			for (uint32_t i = 0; i < vec.size(); i++)
			{
				c_str[i] = InvMapChar(vec[i]);
			}

			c_str[vec.size()] = 0;

			string ret = c_str;

			delete[] c_str;

			return ret;
		}

		vector<uint32_t> StringMapper::MapString(string s)
		{
			vector<uint32_t> ret;

			for (uint32_t i = 0; i < s.length(); i++)
			{
				ret.push_back(MapChar(s[i]));
			}

			return ret;

		}

		uint32_t StringMapper::NumLetters()
		{
			return punctuation_start_index;
		}

		string StringMapper::ConvertToISO(string input)
		{
			char* output_c = new char[input.length() + 1];

			uint32_t next_out_position = 0;
			uint32_t i = 0;
			while (i < input.length())
			{
				if (i < input.length() - 1)
				{
					map<pair<char, char>, char >::iterator it = utf8_iso_map.find(pair<char, char>(input[i], input[i + 1]));
					if (it != utf8_iso_map.end())
					{
						output_c[next_out_position] = it->second;
						next_out_position++;
						i += 2;
						continue;
					}
				}

				output_c[next_out_position] = input[i];
				next_out_position++;
				i++;
			}

			output_c[next_out_position] = 0;

			string output = output_c;

			delete[] output_c;
			return output;
		}

		string StringMapper::ConvertBackFromISO(string iso_str)
		{
			char* output_c = new char[iso_str.length() * 2 + 1];

			uint32_t out_pos = 0;

			for (uint32_t i = 0; i < iso_str.length(); i++)
			{
				map<char, pair<char, char> >::iterator it = iso_utf8_map.find(iso_str[i]);

				if (it != iso_utf8_map.end())
				{
					output_c[out_pos] = it->second.first;
					output_c[out_pos + 1] = it->second.second;
					out_pos += 2;
				}
				else
				{
					output_c[out_pos] = iso_str[i];
					out_pos++;
				}
			}

			output_c[out_pos] = 0;
			string output = output_c;

			delete[] output_c;

			return output;
		}

		string StringMapper::RemoveDiacritics_ISO(string text)
		{

			for (uint32_t i = 0; i < text.length(); i++)
			{
				if (CharIsLetter(text[i]))
				{
					uint32_t letter = MapChar(text[i]);

					if (LetterIsDiacritic(letter))
					{
						text[i] = InvMapChar(GetBasicLetterForDiacriticLetter(letter));
					}

				}
			}


			return text;
		}

		bool StringMapper::CaseInsensitiveIsEqual(string str1, string str2)
		{
			if (str1.length() != str2.length())
				return false;

			for (uint32_t i = 0; i < str1.length(); i++)
			{
				if (CharIsUpperCase(str1[i]))
					str1[i] = GetLowerCaseForUpperCase_char(str1[i]);
				if (CharIsUpperCase(str2[i]))
					str2[i] = GetLowerCaseForUpperCase_char(str2[i]);
			}

			return str1 == str2;
		}


	//StringMapper stringMapper = StringMapper();


}
