/*
 
 Copyright (C) 2010 Michal Richter (michalisek@gmail.com)
 
 This file is part of michalisekSpell.
 
 MichalisekSpell is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 3 of the License, or (at your
 option) any later version.
 
 MichalisekSpell is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with michalisekSpell; see the file COPYING.  If not, see
 <http://www.gnu.org/licenses/>.
 
 */

#include "TransitionCostComputation.hpp"
#include "GlobalVars.hpp"
#include "LMServer.hpp"
#include "NGram.hpp"
#include "ViterbiState.hpp"
#include "StagePosibility.hpp"

namespace ngramchecker {

	double TransitionCostComputation_FormsOnly::ComputeTransitionCost(ViterbiStateP state, StagePosibilityP next)
	{
		uint32_t order = state->history.size() + 1;
		uint32_t *ngram_key_ids = new uint32_t[order];

		StagePosibility_FormsOnlyP nxt = boost::static_pointer_cast<StagePosibility_FormsOnly, StagePosibility>(next);

		ngram_key_ids[0] = nxt->ID();

		for (uint32_t i = 1; i <= state->history.size(); i++)
		{
			StagePosibility_FormsOnlyP h = boost::static_pointer_cast<StagePosibility_FormsOnly, StagePosibility>(state->history[state->history.size() - i]);
			ngram_key_ids[i] = h->ID();
		}

		NGramP ngram_key = NGram::createNGramSearchKey(order, ngram_key_ids);

		double ret_val = form_weight * GlobalVars::lmServer->GetNextWordCost(0, ngram_key);

		//cerr << "prob = " << ret_val << endl << endl;

		return ret_val;


	}

	double TransitionCostComputation_FormsOnly::ComputeTransitionCostSPSequence(vector<StagePosibilityP> &sp_vec, uint32_t start_index, uint32_t end_index)
	{
		uint32_t order = end_index - start_index + 1;
		uint32_t *ngram_key_ids = new uint32_t[order];

		uint32_t index = 0;
		for (int i = (int)end_index; i >= (int)start_index; i--)
		{
			StagePosibility_FormsOnlyP sp = boost::static_pointer_cast<StagePosibility_FormsOnly, StagePosibility>(sp_vec[i]);
			ngram_key_ids[index] = sp->ID();
			index++;
		}

		NGramP ngram_key = NGram::createNGramSearchKey(order, ngram_key_ids);

		return GlobalVars::lmServer->GetNextWordCost(0, ngram_key);
	}


	double TransitionCostComputation_MultiFactor::ComputeTransitionCost(ViterbiStateP state, StagePosibilityP next)
	{
		uint32_t order = state->history.size() + 1;

		uint32_t form_order = min(order, form_lm_order);
		uint32_t lemma_order = min(order, lemma_lm_order);
		uint32_t tag_order = min(order, tag_lm_order);

		uint32_t *form_key_ids = new uint32_t[form_order];
		uint32_t *lemma_key_ids = new uint32_t[lemma_order];
		uint32_t *tag_key_ids = new uint32_t[tag_order];

		StagePosibility_MultiFactorP nxt = boost::static_pointer_cast<StagePosibility_MultiFactor, StagePosibility>(next);

		form_key_ids[0] = nxt->FormID();
		lemma_key_ids[0] = nxt->LemmaID();
		tag_key_ids[0] = nxt->TagID();

		for (uint32_t i = 1; i <= state->history.size(); i++)
		{
			StagePosibility_MultiFactorP h = boost::static_pointer_cast<StagePosibility_MultiFactor, StagePosibility>(state->history[state->history.size() - i]);

			if (i < form_order)
				form_key_ids[i] = h->FormID();

			if (i < lemma_order)
				lemma_key_ids[i] = h->LemmaID();

			if (i < tag_order)
				tag_key_ids[i] = h->TagID();
		}

		NGramP form_ngram_key = NGram::createNGramSearchKey(form_order, form_key_ids);
		NGramP lemma_ngram_key = NGram::createNGramSearchKey(lemma_order, lemma_key_ids);
		NGramP tag_ngram_key = NGram::createNGramSearchKey(tag_order, tag_key_ids);


		return form_weight * GlobalVars::lmServer->GetNextWordCost(0, form_ngram_key) +
			   lemma_weight * GlobalVars::lmServer->GetNextWordCost(1, lemma_ngram_key) +
			   tag_weight * GlobalVars::lmServer->GetNextWordCost(2, tag_ngram_key);

	}

	double TransitionCostComputation_MultiFactor::ComputeTransitionCostSPSequence(vector<StagePosibilityP> &sp_vec, uint32_t start_index, uint32_t end_index)
	{
		uint32_t order = end_index - start_index + 1;

		uint32_t form_order = min(order, form_lm_order);
		uint32_t lemma_order = min(order, lemma_lm_order);
		uint32_t tag_order = min(order, tag_lm_order);

		uint32_t *form_key_ids = new uint32_t[form_order];
		uint32_t *lemma_key_ids = new uint32_t[lemma_order];
		uint32_t *tag_key_ids = new uint32_t[tag_order];

		uint32_t index = 0;
		for (int i = (int)end_index; i >= (int)start_index; i--)
		{
			StagePosibility_MultiFactorP sp = boost::static_pointer_cast<StagePosibility_MultiFactor, StagePosibility>(sp_vec[i]);

			if (index < form_order)
				form_key_ids[index] = sp->FormID();

			if (index < lemma_order)
				lemma_key_ids[index] = sp->LemmaID();

			if (index < tag_order)
				tag_key_ids[index] = sp->TagID();

			index++;
		}

		NGramP form_ngram_key = NGram::createNGramSearchKey(form_order, form_key_ids);
		NGramP lemma_ngram_key = NGram::createNGramSearchKey(lemma_order, lemma_key_ids);
		NGramP tag_ngram_key = NGram::createNGramSearchKey(tag_order, tag_key_ids);

		return form_weight * GlobalVars::lmServer->GetNextWordCost(0, form_ngram_key) +
			   lemma_weight * GlobalVars::lmServer->GetNextWordCost(1, lemma_ngram_key) +
			   tag_weight * GlobalVars::lmServer->GetNextWordCost(2, tag_ngram_key);
	}


	double TransitionCostComputation_MorphoAnal::ComputeTransitionCost(ViterbiStateP state, StagePosibilityP next)
	{
		uint32_t order = state->history.size() + 1;

		uint32_t *lemma_key_ids = new uint32_t[order];
		uint32_t *tag_key_ids = new uint32_t[order];

		StagePosibility_MorphoAnalP nxt = boost::static_pointer_cast<StagePosibility_MorphoAnal, StagePosibility>(next);

		lemma_key_ids[0] = nxt->LemmaID();
		tag_key_ids[0] = nxt->TagID();

		for (uint32_t i = 1; i <= state->history.size(); i++)
		{
			StagePosibility_MorphoAnalP h = boost::static_pointer_cast<StagePosibility_MorphoAnal, StagePosibility>(state->history[state->history.size() - i]);
			lemma_key_ids[i] = h->LemmaID();
			tag_key_ids[i] = h->TagID();
		}

		NGramP lemma_ngram_key = NGram::createNGramSearchKey(order, lemma_key_ids);
		NGramP tag_ngram_key = NGram::createNGramSearchKey(order, tag_key_ids);


		return lemma_weight * GlobalVars::lmServer->GetNextWordCost(1, lemma_ngram_key) +
			   tag_weight * GlobalVars::lmServer->GetNextWordCost(2, tag_ngram_key);
	}

	double TransitionCostComputation_MorphoAnal::ComputeTransitionCostSPSequence(vector<StagePosibilityP> &sp_vec, uint32_t start_index, uint32_t end_index)
	{
		uint32_t order = end_index - start_index + 1;

		uint32_t *lemma_key_ids = new uint32_t[order];
		uint32_t *tag_key_ids = new uint32_t[order];

		uint32_t index = 0;
		for (int i = (int)end_index; i >= (int)start_index; i--)
		{
			StagePosibility_MorphoAnalP sp = boost::static_pointer_cast<StagePosibility_MorphoAnal, StagePosibility>(sp_vec[i]);

			lemma_key_ids[index] = sp->LemmaID();
			tag_key_ids[index] = sp->TagID();

			index++;
		}

		NGramP lemma_ngram_key = NGram::createNGramSearchKey(order, lemma_key_ids);
		NGramP tag_ngram_key = NGram::createNGramSearchKey(order, tag_key_ids);

		return lemma_weight * GlobalVars::lmServer->GetNextWordCost(1, lemma_ngram_key) +
			   tag_weight * GlobalVars::lmServer->GetNextWordCost(2, tag_ngram_key);
	}

	double TransitionCostComputation_Letter::ComputeTransitionCost(ViterbiStateP state, StagePosibilityP next)
	{
		uint32_t order = state->history.size() + 1;
		uint32_t *ngram_key_ids = new uint32_t[order];

		StagePosibility_LetterP nxt = boost::static_pointer_cast<StagePosibility_Letter, StagePosibility>(next);

		ngram_key_ids[0] = nxt->LetterID();

		for (uint32_t i = 1; i <= state->history.size(); i++)
		{
			StagePosibility_LetterP h = boost::static_pointer_cast<StagePosibility_Letter, StagePosibility>(state->history[state->history.size() - i]);
			ngram_key_ids[i] = h->LetterID();
		}

		NGramP ngram_key = NGram::createNGramSearchKey(order, ngram_key_ids);

		if (is_name)
		{
			return GlobalVars::lmServer->GetNextLetterCost_names(ngram_key);
		}
		else
		{
			return GlobalVars::lmServer->GetNextLetterCost_common(ngram_key);
		}
	}

	double TransitionCostComputation_Letter::ComputeTransitionCostSPSequence(vector<StagePosibilityP> &sp_vec, uint32_t start_index, uint32_t end_index)
	{
		uint32_t order = end_index - start_index + 1;
		uint32_t *ngram_key_ids = new uint32_t[order];

		uint32_t index = 0;
		for (int i = (int)end_index; i >= (int)start_index; i--)
		{
			StagePosibility_LetterP sp = boost::static_pointer_cast<StagePosibility_Letter, StagePosibility>(sp_vec[i]);
			ngram_key_ids[index] = sp->LetterID();
			index++;
		}

		NGramP ngram_key = NGram::createNGramSearchKey(order, ngram_key_ids);

		if (is_name)
		{
			return GlobalVars::lmServer->GetNextLetterCost_names(ngram_key);
		}
		else
		{
			return GlobalVars::lmServer->GetNextLetterCost_common(ngram_key);
		}
	}
}